import math
import torch
from .sampling import vonmises_quat_rand, quat_mult
# from display import bo_display


def cyclotron_twist_z(N,L,k,kappa,dtype):
    pos = torch.tensor(L).type(dtype)*torch.rand(N,3).type(dtype)
    
    if type(L)==float or type(L)==int:
        qtarget = torch.cat((
            torch.cos((2*math.pi*k/L)*pos[:,2]/2).reshape(N,1),
            torch.sin((2*math.pi*k/L)*pos[:,2]/2).reshape(N,1),
            torch.zeros(N,1).type(dtype),
            torch.zeros(N,1).type(dtype)),dim=1).type(dtype)
    else:
        qtarget = torch.cat((
            torch.cos((2*math.pi*k/L[2])*pos[:,2]/2).reshape(N,1),
            torch.sin((2*math.pi*k/L[2])*pos[:,2]/2).reshape(N,1),
            torch.zeros(N,1).type(dtype),
            torch.zeros(N,1).type(dtype)),dim=1).type(dtype)    
    bo = vonmises_quat_rand(qtarget, torch.tensor([kappa]).type(dtype))
    
    return pos, bo


def inverted_cyclotron_twist_z(N,L,k,kappa,dtype):
    pos = torch.tensor(L).type(dtype)*torch.rand(N,3).type(dtype)
    bo = torch.zeros(N,4).type(dtype)
   
    if type(L)==float or type(L)==int:
        N1 = torch.sum(pos[:,2]<L/2.).int()
        N2 = torch.sum(pos[:,2]>=L/2.).int()
        qtarget1 = torch.cat((
            torch.cos((2*math.pi*k/L)*pos[pos[:,2]<L/2.,2]/2).reshape(N1,1),
            torch.sin((2*math.pi*k/L)*pos[pos[:,2]<L/2.,2]/2).reshape(N1,1),
            torch.zeros(N1,1).type(dtype),
            torch.zeros(N1,1).type(dtype)),dim=1).type(dtype)
        bo[pos[:,2]<L/2.,:] = vonmises_quat_rand(qtarget1, torch.tensor([kappa]).type(dtype))
       
        qtarget2 = torch.cat((
            torch.cos((2*math.pi*k/L)*pos[pos[:,2]>=L/2.,2]/2).reshape(N2,1),
            -torch.sin((2*math.pi*k/L)*pos[pos[:,2]>=L/2.,2]/2).reshape(N2,1),
            torch.zeros(N2,1).type(dtype),
            torch.zeros(N2,1).type(dtype)),dim=1).type(dtype)
        bo[pos[:,2]>=L/2.,:] = vonmises_quat_rand(qtarget2, torch.tensor([kappa]).type(dtype))
    else:
        N1 = torch.sum(pos[:,2]<L[2]/2.).int()
        N2 = torch.sum(pos[:,2]>=L[2]/2.).int()
        qtarget1 = torch.cat((
            torch.cos((2*math.pi*k/L[2])*pos[pos[:,2]<L[2]/2.,2]/2).reshape(N1,1),
            torch.sin((2*math.pi*k/L[2])*pos[pos[:,2]<L[2]/2.,2]/2).reshape(N1,1),
            torch.zeros(N1,1).type(dtype),
            torch.zeros(N1,1).type(dtype)),dim=1).type(dtype)
        bo[pos[:,2]<L[2]/2.,:] = vonmises_quat_rand(qtarget1, torch.tensor([kappa]).type(dtype))
        
        qtarget2 = torch.cat((
            torch.cos((2*math.pi*k/L[2])*pos[pos[:,2]>=L[2]/2.,2]/2).reshape(N2,1),
            -torch.sin((2*math.pi*k/L[2])*pos[pos[:,2]>=L[2]/2.,2]/2).reshape(N2,1),
            torch.zeros(N2,1).type(dtype),
            torch.zeros(N2,1).type(dtype)),dim=1).type(dtype)
        bo[pos[:,2]>=L[2]/2.,:] = vonmises_quat_rand(qtarget2, torch.tensor([kappa]).type(dtype))
           
    return pos, bo

def helical_twist_x(N,L,k,kappa,dtype):
    pos = torch.tensor(L).type(dtype)*torch.rand(N,3).type(dtype)
    
    if type(L)==float or type(L)==int:
        qtarget = torch.cat((
            torch.cos((2*math.pi*k/L)*pos[:,0]/2).reshape(N,1),
            torch.sin((2*math.pi*k/L)*pos[:,0]/2).reshape(N,1),
            torch.zeros(N,1).type(dtype),
            torch.zeros(N,1).type(dtype)),dim=1).type(dtype)
    else:
        qtarget = torch.cat((
            torch.cos((2*math.pi*k/L[0])*pos[:,0]/2).reshape(N,1),
            torch.sin((2*math.pi*k/L[0])*pos[:,0]/2).reshape(N,1),
            torch.zeros(N,1).type(dtype),
            torch.zeros(N,1).type(dtype)),dim=1).type(dtype)    
    bo = vonmises_quat_rand(qtarget, torch.tensor([kappa]).type(dtype))
    
    return pos, bo


def inverted_helical_twist_x(N,L,k,kappa,dtype):
    pos = torch.tensor(L).type(dtype)*torch.rand(N,3).type(dtype)
    bo = torch.zeros(N,4).type(dtype)
    
    if type(L)==float or type(L)==int:
        N1 = torch.sum(pos[:,0]<L/2.).int()
        N2 = torch.sum(pos[:,0]>=L/2.).int()
        qtarget1 = torch.cat((
            torch.cos((2*math.pi*k/L)*pos[pos[:,0]<L/2.,0]/2).reshape(N1,1),
            torch.sin((2*math.pi*k/L)*pos[pos[:,0]<L/2.,0]/2).reshape(N1,1),
            torch.zeros(N1,1).type(dtype),
            torch.zeros(N1,1).type(dtype)),dim=1).type(dtype)
        bo[pos[:,0]<L/2.,:] = vonmises_quat_rand(qtarget1, torch.tensor([kappa]).type(dtype))
       
        qtarget2 = torch.cat((
            torch.cos((2*math.pi*k/L)*pos[pos[:,0]>=L/2.,0]/2).reshape(N2,1),
            -torch.sin((2*math.pi*k/L)*pos[pos[:,0]>=L/2.,0]/2).reshape(N2,1),
            torch.zeros(N2,1).type(dtype),
            torch.zeros(N2,1).type(dtype)),dim=1).type(dtype)
        bo[pos[:,0]>=L/2.,:] = vonmises_quat_rand(qtarget2, torch.tensor([kappa]).type(dtype))
    else:
        N1 = torch.sum(pos[:,0]<L[0]/2.).int()
        N2 = torch.sum(pos[:,0]>=L[0]/2.).int()
        qtarget1 = torch.cat((
            torch.cos((2*math.pi*k/L[0])*pos[pos[:,0]<L[0]/2.,0]/2).reshape(N1,1),
            torch.sin((2*math.pi*k/L[0])*pos[pos[:,0]<L[0]/2.,0]/2).reshape(N1,1),
            torch.zeros(N1,1).type(dtype),
            torch.zeros(N1,1).type(dtype)),dim=1).type(dtype)
        bo[pos[:,0]<L[0]/2.,:] = vonmises_quat_rand(qtarget1, torch.tensor([kappa]).type(dtype))
        
        qtarget2 = torch.cat((
            torch.cos((2*math.pi*k/L[0])*pos[pos[:,0]>=L[0]/2.,0]/2).reshape(N2,1),
            -torch.sin((2*math.pi*k/L[0])*pos[pos[:,0]>=L[0]/2.,0]/2).reshape(N2,1),
            torch.zeros(N2,1).type(dtype),
            torch.zeros(N2,1).type(dtype)),dim=1).type(dtype)
        bo[pos[:,0]>=L[0]/2.,:] = vonmises_quat_rand(qtarget2, torch.tensor([kappa]).type(dtype))
           
    return pos, bo


import sys
import math
import copy
import time
import numpy as np
import torch
from pykeops.torch import LazyTensor
from .sampling import quat_mult
from .kernels import squared_distances_tensor, lazy_interaction_kernel, \
    lazy_morse, lazy_quadratic
from .toolbox import uniform_grid_centroids, volume_ball, \
    block_sparse_reduction_parameters, maximal_eigenvector_dim2
# from pykeops.torch.cluster import cluster_ranges
from pykeops.torch.cluster import sort_clusters

# from pykeops.torch.cluster import from_matrix

"""Particles as Python iterators."""


class Particles:
    r"""The main class.

    Attributes:
        N (int): Number of particles.
        d (int): Dimension.
        pos ((N,d) Tensor): Positions of the particles.
        R (float or (N,) Tensor): Radius of the particles.
        L ((d,) Tensor): Box size.
        angle (float): Vision angle.
        axis_is_none (bool): True when an axis is specified.
        axis ((N,d) Tensor or None): Axis of the particles.
        bc (list or str): Boundary conditions. Can be
            one of the following:

            - list of size :math:`d` containing 0 (periodic) or 1
              (wall with reflecting boundary conditions) for each
              dimension.

            - ``"open"`` : no boundary conditions.

            - ``"periodic"`` : periodic boundary conditions.

            - ``"spherical"`` : reflecting boundary conditions on the
              sphere of radius ``L[0]/2`` and center **L**/2.

        target_method (dict): Dictionary of target methods (see the
            function :func:`compute_target`).
        target_option_method (dict): Dictionary of options for the
            targets (see the function :func:`compute_target`).
        blocksparse (bool): Use block sparse reduction or not.
        iteration (int): Iteration.
        centroids (Tensor): Centroids.
        eps (Tensor): Size of the cells.
        keep (BoolTensor): ``keep[i,j]`` is True when the cells ``i``
            and ``j`` are contiguous.

    """

    def __init__(self,
                 pos,
                 interaction_radius,
                 box_size,
                 vision_angle=2 * math.pi,
                 axis=None,
                 boundary_conditions='periodic',
                 block_sparse_reduction=False,
                 number_of_cells=1024):
        r"""Initialise the attributes of the class.

        Args:
            pos ((N,d) Tensor): Positions.
            interaction_radius (float or (N,) Tensor): Radius.
            box_size (float or list of size d): Box size.
            vision_angle (float, optional): Vision angle. Default is
                2*pi.
            axis ((N,d) Tensor or None, optional): Axis of the
                particles. Default is None.
            boundary_conditions (list or str, optional): Boundary conditions.
                Can be one of the following:

                - list of size :math:`d` containing 0 (periodic) or 1
                  (wall with reflecting boundary conditions) for each
                  dimension.

                - ``"open"`` : no boundary conditions.

                - ``"periodic"`` : periodic boundary conditions.

                - ``"spherical"`` : reflecting boundary conditions on the
                  sphere of radius ``L[0]/2`` and center **L**/2.

                Default is ``"periodic"``.
            block_sparse_reduction (bool, optional): Use block sparse
                reduction or not. Default is False.
            number_of_cells (int, optional): Maximum number of cells if
                **block_sparse_reduction** is True. Will be rounded to
                the nearest lower power :attr:`d`. Default is 1024.

        """
        self.N, self.d = list(pos.size())
        self.pos = pos
        self.R = interaction_radius
        self.L = torch.tensor(box_size, dtype=pos.dtype, device=pos.device)
        if self.L.shape == torch.Size([]):
            self.L = torch.repeat_interleave(self.L, self.d)
        self.angle = vision_angle
        self.axis_is_none = axis is None
        self.axis = axis
        self.bc = boundary_conditions
        self.target_method = {}
        self.add_target_method("morse", self.morse_target)
        self.add_target_method("quadratic_potential",
                               self.quadratic_potential_target)
        self.add_target_method("overlapping_repulsion",
                               self.overlapping_repulsion_target)
        self.target_option_method = {}
        self.blocksparse = block_sparse_reduction
        (self.eps,
         self.keep,
         self.centroids) = self.compute_blocksparse_parameters(number_of_cells)
        self.iteration = 0

    def __iter__(self):
        return self

    def __next__(self):
        info = self.update()
        self.iteration += 1
        return info

    def update(self):
        raise NotImplementedError()

    def compute_blocksparse_parameters(self, number_of_cells):
        r"""Compute the centroids, the size of the cells and the keep
        matrix.

        Args:
            number_of_cells (int): The maximal number of cells. It will
                be rounded to the nearest lower power :math:`d`.

        Returns:

            (nb_cells, d) Tensor, (nb_cells,) Tensor,
            (nb_cells, nb_cells) BoolTensor:

            The centroids the size of the cells and the keep matrix.

        """
        if self.blocksparse:
            if self.d == 1 or self.d > 3:
                raise NotImplementedError()
            else:
                if isinstance(self.R, float):
                    nbox = (self.L / self.R).floor()
                else:
                    nbox = (self.L / self.R.min()).floor()
                if torch.prod(nbox) <= number_of_cells:
                    eps = self.L / nbox
                    centroids = uniform_grid_centroids(eps, self.L, self.d)
                    D = squared_distances_tensor(
                        centroids,
                        centroids,
                        self.L,
                        boundary_conditions=self.bc
                    )
                    keep = D < ((1.05 * eps) ** 2).sum()
                else:
                    eps0 = (self.L.prod() / number_of_cells) ** (1. / self.d)
                    while eps0 < self.R:
                        eps0 = 1.05 * eps0
                    nbox = (self.L / eps0).floor()
                    eps = self.L / nbox
                    centroids = uniform_grid_centroids(eps, self.L, self.d)
                    D = squared_distances_tensor(
                        centroids,
                        centroids,
                        self.L,
                        boundary_conditions=self.bc
                    )
                    keep = D < ((1.05 * eps) ** 2).sum()
        else:
            eps, keep, centroids = None, None, None
        return eps, keep, centroids

    def best_blocksparse_parameters(self, min_nb_cells,
                                    max_nb_cells, step=1, nb_calls=100,
                                    loop=1, set_best=True):
        r"""Returns the optimal block sparse parameters.

        This function measures the time to execute **nb_calls**
        calls to the :meth:`__next__` for different values of the block
        sparse parameters given by the number of cells to the power
        :math:`1/d`.

        Args:
            min_nb_cells (int): Minimal number of cells to the power
                :math:`1/d`.
            max_nb_cells (int): Maximal number of cells to the power
                :math:`1/d`.
            step (int, optional): Step between two numbers of cells.
                Default is 1.
            nb_calls (int, optional): Number of calls to the function
                :meth:`__next__` at each test. Default is 100.
            loop (int, optional): Number of loops. Default is 1.
            set_best (bool, optional): Set the block sparse parameters
                to the optimal ones. Default is True.

        Returns:
            tuple:

            * **fastest** `(int)`: The number of cells which gave the
                fastest result.
            * **nb_cells** `(list)`: The number of cells tested.
            * **average_simu_time** `(list)`: The average simulation
                times.
            * **simulation_time** `(numpy array)`: The total simulation
                times.

        """
        nb_cells = list(np.arange(min_nb_cells, max_nb_cells, step))
        simulation_time = np.zeros((loop, len(nb_cells)))
        for n, nb in enumerate(nb_cells):
            for i in range(loop):
                new_copy = copy.deepcopy(self)
                new_copy.blocksparse = True
                (new_copy.eps,
                 new_copy.keep,
                 new_copy.centroids) = self.compute_blocksparse_parameters(
                    nb ** new_copy.d)
                start = time.time()
                for k in range(nb_calls):
                    new_copy.__next__()
                end = time.time()
                simulation_time[i, n] = end-start
            progress = round(100*n/len(nb_cells), 2)
            sys.stdout.write('\r' + "Progress:" + str(progress) + "%")
        average_simu_time = np.mean(simulation_time, axis=0)
        fastest = nb_cells[np.argmin(average_simu_time)]
        if set_best:
            self.blocksparse=True
            (self.eps,
             self.keep,
             self.centroids) = self.compute_blocksparse_parameters(
                fastest ** self.d)
        return fastest, nb_cells, average_simu_time, simulation_time

    def add_target_method(self, name, method):
        r"""Add a method to the dictionary :attr:`target_method`.

        Args:
            name (str): The name of the method.
            method (func): A method.

        """
        self.target_method[name] = method

    def add_target_option_method(self, name, method):
        r"""Add an option to the dictionary :attr:`target_option_method`

        Args:
            name (str): The name of the option.
            method (func): An option.

        """
        self.target_option_method[name] = method

    def compute_target(self, which_target, which_options, **kwargs):
        r"""Compute a target and apply some options.

        Args:
            which_target (dict): Dictionary of the form::

                    {"name"       : "method_name",
                     "parameters" : {"param1" : param1,
                                     "param2" : param2,
                                      ...
                                    }
                    }

                The sub-dictionary ``"parameters"`` (eventually empty)
                contains the keyword arguments to be passed to the
                function found in the dictionary :attr:`target_method`
                with the keyword "method_name".
            which_options (dict): Dictionary of the form::

                    {"name_of_option1" : parameters_of_option1,
                     "name_of_option2" : parameters_of_option2,
                    ...
                    }

                The dictionary **`parameters_of_option1`** contains
                the keyword arguments to be passed to the method found
                in the dictionary :attr:`target_option_method` with the
                keyword "name_of_option1".
            **kwargs: Keywords arguments to be passed to the functions
                which compute the target and the options.

        Returns:
            Tensor:

            Compute a target using **which_target** and
            then apply the options in **which_options**
            successively.

        """
        target = self.target_method[which_target["name"]](
            **which_target["parameters"],
            **kwargs)
        if bool(which_options):
            for option_name in which_options:
                target = self.target_option_method[option_name](
                    target,
                    **which_options[option_name],
                    **kwargs)
        return target

    def linear_local_average(self, *to_average,
                             who=None, with_who=None,
                             isaverage=True,
                             kernel=lazy_interaction_kernel):
        r"""Fast computation of a linear local average.

        A linear local average is a matrix vector product between an
        interaction kernel (LazyTensor of size (M,N)) and a Tensor
        of size (N,dimU).

        Args:
            *to_average (Tensor): The quantities to average. The first
                dimension must be equal to :attr:`N`.
            who ((N,) BoolTensor or None, optional): The rows of
                the interaction kernel is ``pos[who,:]`` (or
                :attr:`pos` if **who** is None). Default is None.
            with_who ((N,) BoolTensor or None, optional): The
                columns of the interaction kernel is
                ``pos[with_who,:]`` (or :attr:`pos` if **with_who** is
                None). Default is None.
            isaverage (bool, optional): Divide the result by :attr:`N`
                if **isaverage** is True. Default is True.
            kernel (func, optional): The function which returns the
                interaction kernel as a LazyTensor. Default is
                :func:`lazy_interaction_kernel`.

        Returns:
            tuple of Tensors:

            The linear local averages.

        """
        if who is None:
            who = torch.ones(self.N, dtype=torch.bool, device=self.pos.device)
        if with_who is None:
            with_who = torch.ones(self.N,
                                  dtype=torch.bool, device=self.pos.device)

        x = self.pos[who, :]
        y = self.pos[with_who, :]
        Rx = self.R[who] if type(x) == type(self.R) else self.R
        Ry = self.R[with_who] if type(y) == type(self.R) else self.R
        if self.axis is None:
            axis = None
        else:
            axis = self.axis[who, :]

        M, D = list(x.shape)
        N, D = list(y.shape)

        U = torch.cat(to_average, dim=1)
        dimensions = []
        for u in to_average:
            dimensions.append(u.shape[1])

        dimU = U.shape[1]
        U = U[with_who, :]

        if self.blocksparse:

            (x_sorted,
             y_sorted,
             nb_centro,
             labels_x,
             labels_y,
             permutation_x,
             ranges_ij,
             map_dummies) = block_sparse_reduction_parameters(
                x, y, self.centroids, self.eps, self.L, self.keep,
                where_dummies=False)

            if type(self.R) == type(x):
                Rx_plus = torch.cat((Rx, torch.zeros(nb_centro,
                                                     dtype=Rx.dtype,
                                                     device=Rx.device)))
                Rx_perm = Rx_plus[permutation_x]
                Ry_plus = torch.cat((Ry, torch.zeros(nb_centro,
                                                     dtype=Ry.dtype,
                                                     device=Ry.device)))
                Ry_perm, lryp = sort_clusters(Ry_plus, labels_y)
            else:
                Rx_perm = Rx
                Ry_perm = Ry

            U_plus = torch.cat((U, torch.zeros(nb_centro, dimU,
                                               dtype=U.dtype,
                                               device=U.device)))
            U_sorted, labels_u = sort_clusters(U_plus, labels_y)

            if self.angle < 2 * math.pi:
                dummy_axes = torch.randn((nb_centro, self.d),
                                         dtype = axis.dtype,
                                         device=axis.device)
                dummy_axes = dummy_axes / torch.norm(
                    dummy_axes, dim=1).reshape((nb_centro, 1))
                axis = torch.cat((axis, dummy_axes))
                axis = axis[permutation_x, :]

            K_ij = kernel(x=x_sorted, y=y_sorted,
                          Rx=Rx_perm, Ry=Ry_perm,
                          L=self.L, boundary_conditions=self.bc,
                          vision_angle=self.angle, axis=axis)

            K_ij.ranges = ranges_ij

            J_perm = K_ij @ U_sorted
            J_plus = J_perm[torch.argsort(permutation_x), :]
            # Remove the dummy particles
            J = J_plus[0:M, :]
            if isaverage:
                J = (1. / N) * J

        else:
            K_ij = kernel(x=x, y=y,
                          Rx=Rx, Ry=Ry,
                          L=self.L, boundary_conditions=self.bc,
                          vision_angle=self.angle, axis=axis)

            J = K_ij @ U
            if isaverage:
                J = (1. / self.N) * J
        return torch.split(J, dimensions, dim=1)

    def nonlinear_local_average(self, binary_formula,
                                arg1, arg2,
                                who=None, with_who=None,
                                isaverage=True,
                                kernel=lazy_interaction_kernel):
        r"""Fast computation of a nonlinear local average.

        A nonlinear local average is a reduction of the form

        .. math::
            \sum_j K_{ij} U_{ij}

        where :math:`K_{ij}` is an interaction kernel (
        LazyTensor of size (M,N)) and :math:`U_{ij}` is a
        LazyTensor of size (M,N,dim) computed using a binary
        formula.

        Args:
            binary_formula (func): Takes two arguments **arg1**
                and **arg2** and returns a LazyTensor of size
                (M,N,dim).
            arg1 ((M,D1) Tensor): First argument of the binary formula.
            arg2 ((N,D2) Tensor): Second argument of the binary
                formula.
            who ((N,) BoolTensor or None, optional): The rows of
                the interaction kernel is ``pos[who,:]`` (or
                :attr:`pos` if **who** is None). Default is None.
            with_who ((N,) BoolTensor or None, optional): The
                columns of the interaction kernel is
                ``pos[with_who,:]`` (or :attr:`pos` if **with_who** is
                None). Default is None.
            isaverage (bool, optional): Divide the result by :attr:`N`
                if isaverage is True. Default is True.
            kernel (func, optional): The function which returns the
                interaction kernel as a LazyTensor. Default is
                :func:`lazy_interaction_kernel`.

        Returns:
            (M,dim) Tensor:

            The nonlinear local average.

        """
        if who is None:
            who = torch.ones(self.N, dtype=torch.bool, device=self.pos.device)
        if with_who is None:
            with_who = torch.ones(self.N,
                                  dtype=torch.bool, device=self.pos.device)

        x = self.pos[who, :]
        y = self.pos[with_who, :]
        Rx = self.R[who] if type(x) == type(self.R) else self.R
        Ry = self.R[with_who] if type(y) == type(self.R) else self.R
        if self.axis is None:
            axis = None
        else:
            axis = self.axis[who, :]

        M, D = list(x.shape)
        N, D = list(y.shape)

        if self.blocksparse:

            (x_sorted,
             y_sorted,
             nb_centro,
             labels_x,
             labels_y,
             permutation_x,
             ranges_ij,
             map_dummies) = block_sparse_reduction_parameters(
                x, y, self.centroids, self.eps, self.L, self.keep,
                where_dummies=True)

            if type(self.R) == type(x):
                Rx_plus = torch.cat((Rx, torch.zeros(nb_centro,
                                                     dtype=Rx.dtype,
                                                     device=Rx.device)))
                Rx_perm = Rx_plus[permutation_x]
                Ry_plus = torch.cat((Ry, torch.zeros(nb_centro,
                                                     dtype=Ry.dtype,
                                                     device=Ry.device)))
                Ry_perm, lryp = sort_clusters(Ry_plus, labels_y)
            else:
                Rx_perm = Rx
                Ry_perm = Ry

            if self.angle < 2 * math.pi:
                dummy_axes = torch.randn((nb_centro, self.d),
                                         dtype = axis.dtype,
                                         device=axis.device)
                dummy_axes = dummy_axes / torch.norm(
                    dummy_axes, dim=1).reshape((nb_centro, 1))
                axis = torch.cat((axis, dummy_axes))
                axis = axis[permutation_x, :]

            K_ij = kernel(x=x_sorted, y=y_sorted,
                          Rx=Rx_perm, Ry=Ry_perm,
                          L=self.L, boundary_conditions=self.bc,
                          vision_angle=self.angle, axis=axis)

            N1, D1 = list(arg1.size())
            N2, D2 = list(arg2.size())
            arg1_plus = torch.cat((arg1, torch.zeros(nb_centro, D1,
                                                     dtype=arg1.dtype,
                                                     device=arg1.device)))
            arg2_plus = torch.cat((arg2, torch.zeros(nb_centro, D2,
                                                     dtype=arg2.dtype,
                                                     device=arg2.device)))
            arg1_plus_sorted, la1 = sort_clusters(arg1_plus, labels_x)
            arg2_plus_sorted, la2 = sort_clusters(arg2_plus, labels_y)
            U_ij_plus = map_dummies * binary_formula(arg1_plus_sorted,
                                                     arg2_plus_sorted)

            KU = K_ij * U_ij_plus
            KU.ranges = ranges_ij
            J_perm = KU.sum(1)
            J_plus = J_perm[torch.argsort(permutation_x), :]
            # Remove the dummy particles
            J = J_plus[0:M, :]
            if isaverage:
                J = (1. / N) * J

        else:
            K_ij = kernel(x=x, y=y,
                          Rx=Rx, Ry=Ry,
                          L=self.L, boundary_conditions=self.bc,
                          vision_angle=self.angle, axis=axis)
            U_ij = binary_formula(arg1, arg2)
            KU = K_ij * U_ij
            J = KU.sum(1)
            if isaverage:
                J = (1. / N) * J

        return J

    def number_of_neighbours(self, who=None, with_who=None):
        r""" Compute the number of neighbours.

        The number of neighbours of ``x[i,:]`` is the number of ones in
        the row corresponding to ``x[i,:]`` in the interaction kernel.

        Args:
            who ((N,) BoolTensor or None, optional): The rows of
                the interaction kernel is ``pos[who,:]`` (or
                :attr:`pos` if **who** is None). Default is None.
            with_who ((N,) BoolTensor or None, optional): The
                columns of the interaction kernel is
                ``pos[with_who,:]`` (or :attr:`pos` if **with_who** is
                None). Default is None.

        Returns:
            (M,) Tensor:

            The number of neighbours.

        """
        Nneigh, = self.linear_local_average(
            torch.ones((self.N, 1),
                       dtype=self.pos.dtype, device=self.pos.device),
            who=who, with_who=with_who, isaverage=False)
        return Nneigh.reshape(Nneigh.shape[0])

    def morse_target(self, Ca, la, Cr, lr, p=2, mass=1.,
                     who=None, with_who=None,
                     local=False,
                     isaverage=True,
                     kernel=lazy_interaction_kernel, **kwargs):
        r"""Compute the force exerted on each particle due to the Morse
        potential.

        Args:
            Ca (float): Attraction coefficient.
            la (float): Attraction length.
            Cr (float): Repulsion coefficient.
            lr (float): Repulsion length.
            p (int, optional): Exponent. Default is 2.
            mass ((N,) Tensor or float, optional): Mass of the
                particles. Default is 1.
            who ((N,) BoolTensor or None, optional): The rows of
                the interaction kernel is ``pos[who,:]`` (or
                :attr:`pos` if **who** is None). Default is None.
            with_who ((N,) BoolTensor or None, optional): The
                columns of the interaction kernel is
                ``pos[with_who,:]`` (or :attr:`pos` if **with_who** is
                None). Default is None.
            local (bool, optional): If **local** is true then the sum
                only considers the y particles within the interaction
                kernel. Default is False.
            isaverage (bool, optional): If isaverage is True then divide
                the total force by N where N is the number of True in
                :Tensor"`with_who`. Default is True.
            kernel (func, optional): The interaction kernel to use.
                Default is :func:`lazy_interaction_kernel`
            **kwargs: Arbitrary keyword arguments.

        Returns:
            Tensor:

            The force exerted on each particle located at ``pos[who,:]``.

        """
        if who is None:
            who = torch.ones(self.N, dtype=torch.bool, device=self.pos.device)
        if with_who is None:
            with_who = torch.ones(self.N,
                                  dtype=torch.bool, device=self.pos.device)
        x = self.pos[who, :]
        y = self.pos[with_who, :]
        mx = mass[who] if type(x) == type(mass) else mass
        my = mass[with_who] if type(y) == type(mass) else mass
        if not local:
            lazy_morse_tensor = lazy_morse(x=x, y=y, Ca=Ca, la=la,
                                           Cr=Cr, lr=lr, p=p, mx=mx, my=my)
            if isaverage:
                return (1. / y.shape[0]) * lazy_morse_tensor.sum(axis=1)
            else:
                return lazy_morse_tensor.sum(axis=1)
        else:
            binary_formula = lambda x, y: lazy_morse(x=x, y=y, Ca=Ca, la=la,
                                                     Cr=Cr, lr=lr, p=p,
                                                     mx=mx, my=my)
            return self.nonlinear_local_average(
                binary_formula, x, y,
                who=who, with_who=with_who,
                isaverage=isaverage,
                kernel=kernel)

    def quadratic_potential_target(self, who=None, with_who=None,
                                   kernel=lazy_interaction_kernel,
                                   isaverage=True, **kwargs):
        r"""Compute the force exerted on each particle due to the
        quadratic potential.

        Args:
            who ((N,) BoolTensor or None, optional): The rows of
                the interaction kernel is ``pos[who,:]`` (or
                :attr:`pos` if **who** is None). Default is None.
            with_who ((N,) BoolTensor or None, optional): The
                columns of the interaction kernel is
                ``pos[with_who,:]`` (or :attr:`pos` if **with_who** is
                None). Default is None.
            isaverage (bool, optional): If isaverage is True then divide
                the total force by N where N is the number of True in
                :Tensor"`with_who`. Default is True.
            kernel (func, optional): The interaction kernel to use.
                Default is :func:`lazy_interaction_kernel`
            **kwargs: Arbitrary keyword arguments.

        Returns:
            Tensor:

            The force exerted on each particle located at ``pos[who,:]``.

        """
        if who is None:
            who = torch.ones(self.N, dtype=torch.bool, device=self.pos.device)
        if with_who is None:
            with_who = torch.ones(self.N,
                                  dtype=torch.bool, device=self.pos.device)
        x = self.pos[who, :]
        y = self.pos[with_who, :]
        binary_formula = lambda x, y: lazy_quadratic(
            x=x, y=y, R=self.R, L=self.L, boundary_conditions=self.bc)
        return self.nonlinear_local_average(
            binary_formula, x, y,
            who=who, with_who=with_who,
            isaverage=isaverage,
            kernel=kernel)

    def overlapping_repulsion_target(self, who=None, with_who=None, **kwargs):
        r"""Compute the repulsion force exerted on each particle due to
        the overlapping.

        The force exerted on a particle located at :math:`x_i` derives
        from a logarithmic potential and is given by:

        .. math::

            F = \sum_j K(x_i-x_j) \left(\frac{(R_i+R_j)^2}{|x_i-x_j|^2} - 1\right) (x_i-x_j)

        where :math:`x_j` is the position of the other particles;
        :math:`R_i` and :math:`R_j` are the radii of the particles at
        :math:`x_i` and :math:`x_j` respectively and :math:`K` is the
        overlapping kernel.

        Note:
            in the present case, it is simpler to write explicitly
            the reduction formula but the same function could be
            implemented using a binary formula in
            :func:`nonlinear_local_average` with the kernel given by
            :func:`lazy_overlapping_kernel`.

        Args:
            who ((N,) BoolTensor or None, optional): The rows of
                the interaction kernel is ``pos[who,:]`` (or
                :attr:`pos` if **who** is None). Default is None.
            with_who ((N,) BoolTensor or None, optional): The
                columns of the interaction kernel is
                ``pos[with_who,:]`` (or :attr:`pos` if **with_who** is
                None). Default is None.
            **kwargs: Arbitrary keyword arguments.

        Returns:
            Tensor:

            The force exerted on each particle located at ``pos[who,:]``.

        """
        if who is None:
            who = torch.ones(self.N, dtype=torch.bool, device=self.pos.device)
        if with_who is None:
            with_who = torch.ones(self.N,
                                  dtype=torch.bool, device=self.pos.device)
        x = self.pos[who, :]  # (M,D)
        y = self.pos[with_who, :]  # (N,D)
        Rx = self.R[who] if type(self.R) == type(x) else self.R
        Ry = self.R[with_who] if type(self.R) == type(y) else self.R

        x_i = LazyTensor(x[:, None, :])
        y_j = LazyTensor(y[None, :, :])
        if type(Rx) == type(x):
            R_i = LazyTensor(Rx.reshape((x.shape[0], 1))[:, None])
            R_j = LazyTensor(Ry.reshape((y.shape[0], 1))[None, :])
            R_ij = R_i + R_j  # (M,N)
        else:
            R_ij = Rx + Ry  # float
        z_ij = x_i - y_j  # (M,N,D)
        sq_dist = (z_ij ** 2).sum(-1)  # (M,N)
        R_ij = R_ij * sq_dist.sign()
        sq_dist = sq_dist - (sq_dist.sign() - 1)
        F_ij = (((R_ij ** 2) / sq_dist - 1).relu()) * z_ij

        return F_ij.sum(axis=1)


"""Subclass of kinetic particles."""


class KineticParticles(Particles):
    r"""Subclass of kinetic particles.

    Attributes:
         vel ((N,d) Tensor): Velocities.

    """

    def __init__(self,
                 pos,
                 vel,
                 interaction_radius,
                 box_size,
                 vision_angle=2 * math.pi,
                 axis=None,
                 boundary_conditions='periodic',
                 block_sparse_reduction=False,
                 number_of_cells=1024):
        r"""

        Args:
            pos ((N,d) Tensor): Positions.
            vel ((N,d) Tensor): Velocities.
            interaction_radius (float or (N,) Tensor): Radius.
            box_size (float or list of size d): Box size.
            vision_angle (float, optional): Vision angle. Default is
                :math:`2\pi`.
            axis ((N,d) Tensor or None, optional): Axis of the
                particles. Default is None.
            boundary_conditions (list or str): Boundary conditions.
                Can be one of the following:

                - list of size :math:`d` containing 0 (periodic) or 1
                  (wall with reflecting boundary conditions) for each
                  dimension.

                - ``"open"`` : no boundary conditions.

                - ``"periodic"`` : periodic boundary conditions.

                - ``"spherical"`` : reflecting boundary conditions on the
                  sphere of radius ``L[0]/2`` and center **L**/2.
                Default is ``"periodic"``.
            block_sparse_reduction (bool, optional): Use block sparse
                reduction or not. Default is False.
            number_of_cells (int, optional): Maximum number of cells if
                **block_sparse_reduction** is True. Will be rounded to
                the nearest lower power :attr:`d`. Default is 1024.

        """
        self.vel = vel
        super().__init__(pos=pos,
                         interaction_radius=interaction_radius,
                         box_size=box_size,
                         vision_angle=vision_angle,
                         axis=axis,
                         boundary_conditions=boundary_conditions,
                         block_sparse_reduction=block_sparse_reduction,
                         number_of_cells=number_of_cells)
        self.add_target_method("normalised", self.normalised)
        self.add_target_method("motsch_tadmor", self.motsch_tadmor)
        self.add_target_method("mean_field", self.mean_field)
        self.add_target_method("max_kappa", self.max_kappa)
        self.add_target_method("nematic", self.nematic)
        self.add_target_option_method("bounded_angular_velocity",
                                      self.bounded_angular_velocity)
        self.add_target_option_method("pseudo_nematic", self.pseudo_nematic)

    @property
    def axis(self):
        r"""
        If a None axis is given, then the default axis is the
        normalised velocity.

        Returns:
            (N,d) Tensor

        """
        if self.axis_is_none:
            return self.vel / torch.norm(self.vel, dim=1).reshape((self.N, 1))
        else:
            return self.__axis

    @axis.setter
    def axis(self, value):
        if self.axis_is_none:
            """DO NOTHING"""
        else:
            self.__axis = value

    @property
    def order_parameter(self):
        r"""Compute the norm of the average velocity.

        Returns:
            float:

            The order parameter.

        """
        vec_phi = (1. / self.N) * self.vel.sum(0)
        phi = torch.norm(vec_phi)
        return float(phi)

    def check_boundary(self):
        r"""Update the positions and velocities of the particles which
           are outside the boundary.

        """
        if self.bc == 'periodic':
            for i in range(self.d):
                self.pos[:, i] = torch.remainder(self.pos[:, i], self.L[i])
        elif type(self.bc) == list:
            for i in range(self.d):
                if self.bc[i] == 1:
                    inf0 = self.pos[:, i] < 0
                    supL = self.pos[:, i] > self.L[i]
                    self.pos[inf0, i] = -self.pos[inf0, i]
                    self.pos[supL, i] = 2 * self.L[i] - self.pos[supL, i]

                    self.vel[inf0, i] = -self.vel[inf0, i]
                    self.vel[supL, i] = -self.vel[supL, i]
                elif self.bc[i] == 0:
                    self.pos[:, i] = torch.remainder(self.pos[:, i], self.L[i])
        elif self.bc == 'spherical':
            # Translate the center to the origin and compute the
            # radius = L/2.
            x = self.pos - self.L / 2
            r = self.L[0] / 2
            out = (x ** 2).sum(1) > (r ** 2)
            N_out = out.sum()
            if N_out > 0:
                x_out = x[out, :]
                v_out = self.vel[out, :]
                # Find x0 and c0 such that x = x0 + c0*vel with |x0|=r
                xdotv = (x_out * v_out).sum(1)
                sqnorm_v_out = (v_out ** 2).sum(1)
                sqnorm_x_out = (x_out ** 2).sum(1)
                D0 = (xdotv ** 2
                      - sqnorm_v_out * (sqnorm_x_out - r ** 2)
                      ).abs()  # Take the absolute value to ensure >=0
                c0 = 1. / sqnorm_v_out * (xdotv - D0.sqrt())
                c0 = c0.reshape((N_out, 1))
                x0 = x_out - c0 * v_out
                # Find x1 and c1 such that x = x1+c1*x0 with |x1|=r
                xdotx0 = (x_out * x0).sum(1)
                D1 = (xdotx0 ** 2 - (r ** 2) * (sqnorm_x_out - r ** 2)).abs()
                c1 = (r ** (-2)) * (xdotx0 - D1.sqrt())
                c1 = c1.reshape((N_out, 1))
                # Find c2 such that v = v2+c2*x0 and |v|=|v2|
                c2 = 2. * (r ** (-2)) * ((v_out * x0).sum(1))
                c2 = c2.reshape((N_out, 1))
                # Construct the reflections
                x_reflect = x_out - 2 * c1 * x0
                v_reflect = v_out - c2 * x0
                v_reflect = v_reflect / torch.norm(v_reflect,
                                                   dim=1).reshape((N_out, 1)) \
                            * sqnorm_v_out.sqrt().reshape((N_out, 1))
                # Update the positions and velocity
                self.pos[out, :] = x_reflect + self.L / 2
                self.vel[out, :] = v_reflect
                # Stability and emergency procedure....
                self.pos = torch.max(self.pos,
                                     torch.tensor([0.],
                                                  dtype=self.pos.dtype,
                                                  device=self.pos.device))
                self.pos = torch.min(self.pos, self.L[0])
                x = self.pos - self.L / 2
                r = self.L[0] / 2
                out = (x ** 2).sum(1) > (r ** 2)
                N_out = out.sum()
                if N_out > 0:
                    if N_out > 42:
                        print('Warning!'+str(N_out)+'particles out')
                    self.pos[out, :] = self.L/2 + torch.zeros(
                        (N_out, self.d),
                        dtype=self.pos.dtype,
                        device=self.pos.device
                    )

    def normalised(self, kappa, who=None, with_who=None, **kwargs):
        r"""The normalised average velocity.

        Given a system of particles with positions :math:`(x_i)_i` and
        velocities :math:`(v_i)_i`, the normalised average velocity
        around particle :math:`(x_i,v_i)` is

        .. math::

            \Omega_i = \frac{\sum_j K(x_i-x_j) v_j}{\left|\sum_j K(x_i-x_j) v_j\right|},

        where :math:`K` is the interaction kernel.

        Args:
            kappa ((1,) Tensor or (M,) Tensor): The concentration
                parameter. If the size of kappa is bigger than 1 then
                its size must be equal to the number of True values in
                **who**.
            who ((N,) BoolTensor or None, optional): The rows of
                the interaction kernel is ``pos[who,:]`` (or
                :attr:`pos` if **who** is None). Default is None.
            with_who ((N,) BoolTensor or None, optional): The
                columns of the interaction kernel is
                ``pos[with_who,:]`` (or :attr:`pos` if **with_who** is
                None). Default is None.
            **kwargs: Arbitrary keyword arguments.

        Returns:
            Tensor:

            The normalised averaged velocity multiplied by the
            concentration parameter.

        """
        J, = self.linear_local_average(self.vel, who=who, with_who=with_who)
        Omega = J / torch.norm(J, dim=1).reshape((J.shape[0], 1))
        return kappa * Omega

    def motsch_tadmor(self, kappa, who=None, with_who=None, **kwargs):
        r"""The average velocity of the neighbours only.

        Given a system of particles with positions :math:`(x_i)_i` and
        velocities :math:`(v_i)_i`, the average velocity
        around particle :math:`(x_i,v_i)` with the Motsch-Tadmor
        normalisation is:

        .. math::

            \Omega_i = \frac{\sum_j K(x_i-x_j) v_j}{\sum_j K(x_i-x_j)},

        where :math:`K` is the interaction kernel.

        Args:
            kappa ((1,) Tensor or (M,) Tensor): The concentration
                parameter. If the size of kappa is bigger than 1 then
                its size must be equal to the number of True values in
                **who**.
            who ((N,) BoolTensor or None, optional): The rows of
                the interaction kernel is ``pos[who,:]`` (or
                :attr:`pos` if **who** is None). Default is None.
            with_who ((N,) BoolTensor or None, optional): The
                columns of the interaction kernel is
                ``pos[with_who,:]`` (or :attr:`pos` if **with_who** is
                None). Default is None.
            **kwargs: Arbitrary keyword arguments.

        Returns:
            Tensor:

            The average velocity of the neighbours multiplied
            by the concentration parameter.

        """
        justN = torch.ones((self.N, 1),
                           dtype=self.pos.dtype, device=self.pos.device)
        J, N_neighbours = self.linear_local_average(self.vel, justN,
                                                    who=who, with_who=with_who)
        return kappa * J / N_neighbours

    def mean_field(self, kappa, who=None, with_who=None, **kwargs):
        r"""Non-normalised average velocity.

        Given a system of particles with positions :math:`(x_i)_i` and
        velocities :math:`(v_i)_i`, the mean-field average velocity
        around particle :math:`(x_i,v_i)` (without normalisation) is:

        .. math::

            J_i = \frac{1}{N}\sum_j K(x_i-x_j) v_j,

        where :math:`K` is the interaction kernel.

        Args:
            kappa ((1,) Tensor or (M,) Tensor): The concentration
                parameter. If the size of kappa is bigger than 1 then
                its size must be equal to the number of True values in
                **who**.
            who ((N,) BoolTensor or None, optional): The rows of
                the interaction kernel is ``pos[who,:]`` (or
                :attr:`pos` if **who** is None). Default is None.
            with_who ((N,) BoolTensor or None, optional): The
                columns of the interaction kernel is
                ``pos[with_who,:]`` (or :attr:`pos` if **with_who** is
                None). Default is None.
            **kwargs: Arbitrary keyword arguments.

        Returns:
            Tensor:

            The mean-field average velocity multiplied by a scaling
            parameter and by the concentration parameter. The scaling
            parameter is equal to the volume of the box divided by
            the volume of the ball of radius :attr:`R` in dimension
            :attr:`d`.

        """
        scaling = (self.L / self.R) ** self.d * 1. / volume_ball(self.d)
        J, = self.linear_local_average(self.vel, who=who, with_who=with_who)
        J = scaling * J
        return kappa * J

    def max_kappa(self, kappa, kappa_max, who=None, with_who=None, **kwargs):
        r""" The mean-field target with a norm threshold.

        Given a system of particles with positions :math:`(x_i)_i` and
        velocities :math:`(v_i)_i`, the mean-field average velocity
        around particle :math:`(x_i,v_i)` with a cutoff is:

        .. math::

            \Omega_i = \frac{1}{\frac{1}{\kappa}+\frac{|J_i|}{\kappa_\mathrm{max}}} J_i

        where

        .. math::

            J_i = \frac{1}{N}\sum_j K(x_i-x_j) v_j,

        and :math:`K` is the interaction kernel multiplied by the
        scaling parameter. The scaling parameter is equal to the
        volume of the box divided by the volume of the ball of radius
        :attr:`R` in dimension :attr:`d`. The norm of :math:`\Omega_i`
        is thus bounded by :math:`\kappa_\mathrm{max}`.

        Args:
            kappa ((1,) Tensor or (M,) Tensor): The concentration
                parameter. If the size of kappa is bigger than 1 then
                its size must be equal to the number of True values in
                **who**.
            who ((N,) BoolTensor or None, optional): The rows of
                the interaction kernel is ``pos[who,:]`` (or
                :attr:`pos` if **who** is None). Default is None.
            with_who ((N,) BoolTensor or None, optional): The
                columns of the interaction kernel is
                ``pos[with_who,:]`` (or :attr:`pos` if **with_who** is
                None). Default is None.
            **kwargs: Arbitrary keyword arguments.

        Returns:
            Tensor:

            The cutoff average velocity.

        """
        scaling = (self.L / self.R) ** self.d * 1. / volume_ball(self.d)
        J, = self.linear_local_average(self.vel, who=who, with_who=with_who)
        J = scaling * J
        normJ = torch.norm(J, dim=1).reshape((J.shape[0], 1))
        return 1. / ((1. / kappa) + (normJ / kappa_max)) * J

    def nematic(self, kappa, who=None, with_who=None, **kwargs):
        r"""Normalised average of the nematic velocities.

        Given a system of particles with positions :math:`(x_i)_i` and
        velocities :math:`(v_i)_i`, the nematic average velocity
        around particle :math:`(x_i,v_i)` is:

        .. math::

            \Omega_i = (v_i\cdot \overline{\Omega}_i)\overline{\Omega}_i,

        where :math:`\overline{\Omega}_i` is any unit eigenvector
        associated to the maximal eigenvalue of the average Q-tensor:

        .. math::

            Q_i = \sum_j K(x_i-x_j) \left(v_j\otimes v_j - \frac{1}{d} I_d\right),

        where :math:`K` is the interaction kernel.

        Args:
            kappa ((1,) Tensor or (M,) Tensor): The concentration
                parameter. If the size of kappa is bigger than 1 then
                its size must be equal to the number of True values in
                **who**.
            who ((N,) BoolTensor or None, optional): The rows of
                the interaction kernel is ``pos[who,:]`` (or
                :attr:`pos` if **who** is None). Default is None.
            with_who ((N,) BoolTensor or None, optional): The
                columns of the interaction kernel is
                ``pos[with_who,:]`` (or :attr:`pos` if **with_who** is
                None). Default is None.
            **kwargs: Arbitrary keyword arguments.

        Returns:
            Tensor:

            The nematic average velocity multiplied by the concentration
            parameter.

        """
        if who is None:
            who = torch.ones(self.N, dtype=torch.bool, device=self.pos.device)
        Q = torch.einsum('ijl,ilk->ijk',
                         self.vel[:, :, None], self.vel[:, None, :]) \
            - (1. / self.d) * torch.eye(self.d,
                                        dtype=self.vel.dtype,
                                        device=self.vel.device)
        reshapeQ = Q.reshape(self.N, self.d ** 2)
        J, = self.linear_local_average(reshapeQ, who=who, with_who=with_who)
        M = J.shape[0]
        J = J.reshape(M, self.d, self.d)
        if self.d == 2:
            targets = maximal_eigenvector_dim2(J)
        else:
            eivel, eivec = torch.linalg.eigh(J.to('cpu'))
            targets = eivec[:, :, -1]
            targets = targets.to(J.device)
        dot = (self.vel[who, :] * targets).sum(1).reshape((M, 1))
        targets = dot * targets
        return kappa * targets

    def bounded_angular_velocity(self, target, angvel_max, dt,
                                 who=None, **kwargs):
        r"""Bounded angle between the velocity and the target.

        The maximum angle between the velocity and the new target cannot
        exceed the value ``angvel_max * dt``. The output tensor
        **new_target** is of the form::

            new_target = a * vel + b * target

        where :math:`a,b>0` are such that the cosine between
        **new_target** and :attr:`vel` is equal to::

            max_cos := max(cos(angvel_max * dt), cos_target)

        where **cos_target** is the cosine between **target** and
        :attr:`vel`. The solution is::

            b = (1 - max_cos ** 2) / (1 - cos_target ** 2)
            a = |target| / |vel| * (max_cos - b * cos_target)

        Args:
            target ((M,d) Tensor): Target.
            angvel_max (float): Maximum angular velocity.
            dt (float): Interval of time.
            who ((N,) BoolTensor, optional): The velocities of the
                particles associated to the target are ``vel[who,:]``
                (or :attr:`vel` if **who** is None). Default is None.
            **kwargs: Arbitrary keywords arguments.

        Returns:
            (M,d) Tensor:

            The new target.

        """
        if who is None:
            who = torch.ones(self.N, dtype=torch.bool, device=self.pos.device)
        M = target.shape[0]
        vel = self.vel[who, :]
        norm_vel = torch.norm(vel, dim=1).reshape((M, 1))
        norm_target = torch.norm(target, dim=1).reshape((M, 1))

        cos_max = math.cos(angvel_max * dt)
        dot = (vel * target).sum(axis=1).reshape((M, 1))
        cos = dot / (norm_vel * norm_target)
        max_cos = dot / (norm_vel * norm_target)
        max_cos[max_cos <= cos_max] = cos_max

        # Add .01 for stability
        b = torch.sqrt((1.01 - max_cos ** 2) / (1.01 - cos ** 2))
        a = (norm_target / norm_vel) * (max_cos - b * cos)

        new_target = a * vel + b * target

        return new_target

    def pseudo_nematic(self, target, who=None, **kwargs):
        r"""Choose the closest to the velocity between the
        target and its opposite.

        Args:
            target ((M,d) Tensor): Target.
            who ((N,) BoolTensor, optional): The velocities of the
                particles associated to the target are ``vel[who,:]``
                (or :attr:`vel` if **who** is None). Default is None.
            **kwargs: Arbitrary keywords arguments.

        Returns:
            (M,d) Tensor:

            Take the opposite of **target** if the angle with the
            velocity is larger than :math:`\pi/2`.

        """
        if who is None:
            who = torch.ones(self.N, dtype=torch.bool, device=self.pos.device)
        dot = (self.vel[who, :] * target).sum(axis=1)
        target[dot < 0] = -target[dot < 0]
        return target


"""The subclass of body-oriented particles."""


class BOParticles(Particles):
    r"""3D particle with a body-orientation in :math:`SO(3)`.

    A body-orientation is a rotation matrix is :math:`SO(3)` or
    equivalently a unit quaternion.
    If :math:`q = (x, y, z, t)` is a unit quaternion, the associated
    rotation matrix is :math:`A = [e_1, e_2, e_3]` where the column
    vectors are defined by:

    .. math::

        e_1 = (x^2 + y^2 - z^2 + t^2, 2(xt+yz), 2(yt-xz))^T

    .. math::
        e_2 = (2(yz-xt), x^2 - y^2 + z^2 - t^2, 2(xy+zt))^T

    .. math::
        e_3 = (2(xz+yt), 2(zt-xy), x^2 - y^2 - z^2 + t^2)^T

    Conversely, if :math:`A` is the rotation of angle :math:`\theta`
    around the axis :math:`u = (u_1, u_2, u_3)^T` then the associated
    unit quaternion is:

    .. math::

        q = \cos(\theta/2) + \sin(\theta/2)(iu_1 + ju_2 + ku_3).

    or its opposite.

    Attributes:
        bo ((N,4) Tensor): Body-orientation stored as a unit quaternion.

    """

    def __init__(self, pos, bo,
                 interaction_radius, box_size,
                 vision_angle=2 * math.pi, axis=None,
                 boundary_conditions='periodic',
                 block_sparse_reduction=False, number_of_cells=1024):
        r"""

        Args:
            pos ((N,3) Tensor): Positions.
            bo ((N,4) Tensor): Body-orientations (unit quaternions).
            interaction_radius (float or (N,) Tensor): Radius.
            box_size (float or list of size d): Box size.
            vision_angle (float, optional): Vision angle. Default is
                :math:`2\pi`.
            axis ((N,3) Tensor or None, optional): Axis of the
                particles. Default is None.
            boundary_conditions (list or str, optional): Boundary conditions.
                Can be one of the following:

                - list of size :math:`d` containing 0 (periodic) or 1
                  (wall with reflecting boundary conditions) for each
                  dimension.

                - ``"open"`` : no boundary conditions.

                - ``"periodic"`` : periodic boundary conditions.

                - ``"spherical"`` : reflecting boundary conditions on the
                  sphere of radius ``L[0]/2`` and center **L**/2.

                Default is ``"periodic"``.
            block_sparse_reduction (bool, optional): Use block sparse
                reduction or not. Default is False
            number_of_cells (int, optional): Maximum number of cells if
                **block_sparse_reduction** is True. Will be rounded to
                the nearest lower power :attr:`d`. Default is 1024.

        """
        self.bo = bo
        super().__init__(pos,
                         interaction_radius,
                         box_size,
                         vision_angle=vision_angle, axis=axis,
                         boundary_conditions=boundary_conditions,
                         block_sparse_reduction=block_sparse_reduction,
                         number_of_cells=number_of_cells)
        self.add_target_method("normalised", self.normalised)

    @property
    def vel(self):
        r"""The velocity is the first column vector of the
        body-orientation matrix.

        Returns:
            (N,3) Tensor:

            The velocity.

        """
        x = self.bo[:, 0]
        y = self.bo[:, 1]
        z = self.bo[:, 2]
        t = self.bo[:, 3]
        e1 = torch.stack(
            (x ** 2 + y ** 2 - z ** 2 - t ** 2,
             2 * (x * t + y * z),
             2 * (y * t - x * z)), dim=1)
        return e1

    @property
    def axis(self):
        r""" If a None axis is given, then the default axis is the
        (normalised) velocity.

        Returns:
            (N,3) Tensor:

        """
        if self.axis_is_none:
            return self.vel
        else:
            return self.__axis

    @axis.setter
    def axis(self, value):
        if self.axis_is_none:
            """DO NOTHING"""
        else:
            self.__axis = value

    @property
    def orth_basis(self):
        r"""An othogonal basis of the orthogonal plane of the velocity.

        Returns:
            tuple:

            * **e2** *((N,3) Tensor)*: Second column vector of the
              body-orientation matrix.

            * **e3** *((N,3) Tensor)*: Third column vector of the
              body-orientation matrix.

        """
        x = self.bo[:, 0]
        y = self.bo[:, 1]
        z = self.bo[:, 2]
        t = self.bo[:, 3]
        e2 = torch.stack(
            (2 * (y * z - x * t),
             x ** 2 - y ** 2 + z ** 2 - t ** 2,
             2 * (x * y + z * t)), dim=1)
        e3 = torch.stack(
            (2 * (x * z + y * t),
             2 * (z * t - x * y),
             x ** 2 - y ** 2 - z ** 2 + t ** 2), dim=1)
        return e2, e3

    @property
    def qtensor(self):
        r"""Compute the normalised uniaxial Q-tensors of the particles.

        If :math:`q` is a unit quaternion seen as a colum vector of
        dimension 4, the normalised uniaxial Q-tensor is defined by:

        .. math::

            Q = q \otimes q - \frac{1}{4}I_4

        where :math:`\otimes` is the tensor product and :math:`I_4` the
        identity matrix of size 4.

        Returns:
            (N,4,4) Tensor:

            Q-tensors of the N particles.

        """
        Q = torch.einsum('ijl,ilk->ijk',
                         self.bo[:, :, None], self.bo[:, None, :]) \
            - .25 * torch.eye(4, dtype=self.bo.dtype, device=self.bo.device)
        return Q

    @property
    def omega_basis(self):
        r"""Local frame associated to the direction of motion.

        The unit vector :math:`\Omega` is the direction of motion
        (it is equal to :attr:`vel`) and
        :math:`[\Omega, e_\theta, e_\varphi]` is the local orthomormal
        frame associated to the spherical cooordinates.

        Returns:
            tuple:

            * **p** *((N,3) Tensor)*: :math:`e_\varphi`.
            * **q** *((N,3) Tensor)*: :math:`-e_\theta`.
        """
        Omega = self.vel
        phi = torch.atan2(Omega[:, 1], Omega[:, 0])
        p = torch.cat(
            (-torch.sin(phi).reshape((self.N, 1)),
             torch.cos(phi).reshape((self.N, 1)),
             torch.zeros((self.N, 1), dtype=phi.dtype, device=phi.device)),
            dim=1)
        q = torch.cross(Omega, p)
        return p, q

    @property
    def u_in_omega_basis(self):
        r"""Coordinates of :math:`u` in the frame :math:`[\Omega, p, q]`
        given by :attr:`omega_basis` where :math:`u` is the second
        column vector of the body-orientation matrix."""
        p, q = self.omega_basis
        u, v = self.orth_basis
        up = (u * p).sum(dim=1)
        uq = (u * q).sum(dim=1)
        return up, uq

    def normalised(self, who=None, with_who=None, **kwargs):
        r"""Projection of the average body-orientation on :math:`SO(3)`.

        Given a system of particles with positions :math:`(x_i)_i` and
        body-orientation matrices :math:`(A_i)_i`, the mean-field average
        body-orientation around the particle :math:`(x_i,A_i)` is
        is the arithmetic average:

        .. math::

            J_i = \sum_j K(x_i - x_j) A_j

        where :math:`K` is the interaction kernel. The projection on
        :math:`SO(3)` is

        .. math::

            P_{SO(3)}(J_i) = argmax_{A\in SO(3)}( A \cdot J_i )

        for the dot product

        .. math::

            A \cdot B = \frac{1}{2} Tr(A^T B).

        The unit quaternion associated to :math:`P_{SO(3)}(J_i)` is any
        eigenvector associated to the maximal eigenvalue of the average
        Q-tensor:

        .. math::

            \overline{Q}_i = \sum_j K(x_i - x_j) Q_j

        where :math:`Q_j` is the Q-tensor of particle :math:`(x_j,A_j)`
        given by :attr:`qtensor`.

        Args:
            who ((N,) BoolTensor or None, optional): The rows of
                the interaction kernel is ``pos[who,:]`` (or
                :attr:`pos` if **who** is None). Default is None.
            with_who ((N,) BoolTensor or None, optional): The
                columns of the interaction kernel is
                ``pos[with_who,:]`` (or :attr:`pos` if **with_who** is
                None). Default is None.
            **kwargs: Arbitrary keyword arguments.

        Returns:
            (M,4) Tensor:

            The unit quaternion associated to the projection of the
            mean-field average body-orientation on :math:`SO(3)`.

        """
        reshapeQ = self.qtensor.reshape(self.N, 16)
        J, = self.linear_local_average(reshapeQ, who=who, with_who=with_who)
        M = J.shape[0]
        J = J.reshape(M, 4, 4)
        # Faster on the cpu !
        eivel, eivec = torch.linalg.eigh(J.to('cpu'))
        targets = eivec[:, :, 3]
        # Back to the gpu.
        targets = targets.to(J.device)
        return targets

    ### Some statistical quantities ###

    @property
    def theta(self):
        r"""Polar angle of the direction of motion"""
        Omega = self.vel.sum(0)
        Omega = Omega / torch.norm(Omega)
        return float(torch.acos(Omega[2]))

    @property
    def phi(self):
        r"""Azimuthal angle of the direction of motion"""
        Omega = self.vel.sum(0)
        Omega = Omega / torch.norm(Omega)
        phi_pi = float(torch.atan2(Omega[1], Omega[0]))
        return np.mod(phi_pi, 2 * np.pi)

    def local_vel(self, xmin=0, xmax=1, ymin=0, ymax=1, zmin=0, zmax=1,
                  who=None):
        if who is None:
            index = (self.pos[:, 0] >= xmin and self.pos[:, 0] <= xmax) \
                    and (self.pos[:, 0] >= xmin and self.pos[:, 0] <= xmax) \
                    and (self.pos[:, 0] >= xmin and self.pos[:, 0] <= xmax)
        else:
            index = who
        Omega = self.vel[index, :].sum(0)
        order_parameter = torch.norm(Omega) / len(self.vel[index, 0])
        Omega = Omega / torch.norm(Omega)
        theta = float(torch.acos(Omega[2]))
        phi = np.mod(float(torch.atan2(Omega[1], Omega[0])), 2 * np.pi)
        return order_parameter, theta, phi

    def twist(self, K, axis, return_slice_nb=None, local_vel=False):
        up, uq = self.u_in_omega_basis
        if axis == 'x':
            if self.L.shape == torch.Size([]):
                eps = self.L / K
            else:
                eps = self.L[0] / K
            index = (self.pos[:, 0] / eps).floor().int()
        if axis == 'y':
            if self.L.shape == torch.Size([]):
                eps = self.L / K
            else:
                eps = self.L[1] / K
            index = (self.pos[:, 1] / eps).floor().int()
        if axis == 'z':
            if self.L.shape == torch.Size([]):
                eps = self.L / K
            else:
                eps = self.L[2] / K
            index = (self.pos[:, 2] / eps).floor().int()
        else:
            raise ValueError('axis must be x, y or z')
        index = torch.max(index, torch.tensor(0, device=index.device))
        index = torch.min(index, torch.tensor(K - 1, device=index.device))
        slice_p = torch.zeros(K, dtype=up.dtype, device=up.device)
        slice_q = torch.zeros(K, dtype=uq.dtype, device=uq.device)
        if local_vel:
            op = torch.zeros(K, dtype=up.dtype, device=up.device)
            theta_loc = torch.zeros(K, dtype=up.dtype, device=up.device)
            phi_loc = torch.zeros(K, dtype=up.dtype, device=up.device)
        for k in range(K):
            Nk = torch.sum(index == k)
            slice_p[k] = torch.sum(up[index == k]) / Nk
            slice_q[k] = torch.sum(uq[index == k]) / Nk
            if local_vel:
                who_k = (index == k)
                opk, thk, phk = self.local_vel(who=who_k)
                op[k] = opk
                theta_loc[k] = thk
                phi_loc[k] = phk
        if return_slice_nb is None:
            if local_vel:
                return slice_p, slice_q, op, theta_loc, phi_loc
            else:
                return slice_p, slice_q
        else:
            if local_vel:
                return (slice_p, slice_q,
                        up[index == return_slice_nb],
                        uq[index == return_slice_nb],
                        op, theta_loc, phi_loc)
            else:
                return (slice_p, slice_q,
                        up[index == return_slice_nb],
                        uq[index == return_slice_nb])

    @property
    def local_order(self):
        def alpha(arg1, arg2):
            arg_i = LazyTensor(arg1[:, None, :])
            arg_j = LazyTensor(arg2[None, :, :])
            return ((arg_i * arg_j).sum(-1)) ** 2

        Phi = self.nonlinear_local_average(alpha, self.bo, self.bo)
        Nneigh = self.number_of_neighbours()
        Phi = Phi - (1. / self.N)
        Nneigh = torch.max(Nneigh - (1. / self.N),
                           torch.tensor(1. / self.N,
                                        dtype=Nneigh.dtype,
                                        device=Nneigh.device))
        order_i = Phi * (Nneigh ** (-1))
        return (1 / self.N) * order_i.sum().item()

    @property
    def global_order(self):
        r"""Returns

         .. math::

            \frac{1}{ N(N-1)} \sum_{i,j} (q_i \cdot q_j)^2.

        """
        bo_i = LazyTensor(self.bo[:, None, :])
        bo_j = LazyTensor(self.bo[None, :, :])
        bo_ij = ((bo_i * bo_j).sum(-1)) ** 2
        order_i = 1. / (self.N - 1.) * (bo_ij.sum(1) - 1.)
        return (1. / self.N) * order_i.sum().item()

    @property
    def order_parameter(self):
        return self.global_order

    def build_reflection_quat(self, who, axis):
        r"""

        Returns the quaternion associated to the rotation of angle
        :math:`2\alpha` and axis :math:`n` where :math:`\alpha` is the
        angle between ``vel[who,:]`` and the plane  orthogonal to
        **axis** and :math:`n` is the element of this plane obtained by
        a rotation of :math:`\pi/2` around **axis** of the normalised
        projection of ``vel[who,:]``.

        Args:
            who ((N,) BoolTensor)
            axis (int): The axis number either 0 (:math:`x` axis), 1
                (:math:`y` axis) or 2 (:math:`z` axis).

        Returns:
            (M,4) Tensor:

            Rotation (unit quaternion).

        """
        Omega = self.vel[who, :]
        Om1 = Omega[:, 0]
        Om2 = Omega[:, 1]
        Om3 = Omega[:, 2]
        K = Omega.shape[0]
        if axis == 0:
            q0 = torch.sqrt(Om2 ** 2 + Om3 ** 2)
            q = torch.stack(
                (q0,
                 torch.zeros(K, dtype=self.bo.dtype, device=self.bo.device),
                 -Om1 * Om3 / q0,
                 Om1 * Om2 / q0), dim=1)
        elif axis == 1:
            q0 = torch.sqrt(Om3 ** 2 + Om1 ** 2)
            q = torch.stack(
                (q0,
                 -Om2 * Om3 / q0,
                 torch.zeros(K, dtype=self.bo.dtype, device=self.bo.device),
                 Om2 * Om1 / q0), dim=1)
        elif axis == 2:
            q0 = torch.sqrt(Om1 ** 2 + Om2 ** 2)
            q = torch.stack(
                (q0,
                 -Om3 * Om2 / q0,
                 Om3 * Om1 / q0,
                 torch.zeros(K, dtype=self.bo.dtype, device=self.bo.device)),
                dim=1)
        else:
            raise ValueError('axis must be 0, 1 or 2.')
        return q

    def check_boundary(self):
        r"""Update the positions and body-orientations of the particles
           which are outside the boundary.

        """
        if self.bc == 'periodic':
            for i in range(self.d):
                self.pos[:, i] = torch.remainder(self.pos[:, i], self.L[i])
        elif type(self.bc) == list:
            for i in range(self.d):
                if self.bc[i] == 1:
                    inf0 = self.pos[:, i] < 0
                    supL = self.pos[:, i] > self.L[i]
                    self.pos[inf0, i] = torch.max(
                        -self.pos[inf0, i],
                        torch.tensor([0.],
                                     dtype=self.pos.dtype,
                                     device=self.pos.device))
                    self.pos[supL, i] = torch.min(
                        2 * self.L[i] - self.pos[supL, i],
                        self.L[i])

                    q_infzero = self.build_reflection_quat(who=inf0, axis=i)
                    self.bo[inf0, :] = quat_mult(q_infzero,
                                                 self.bo[inf0, :])
                    q_supL = self.build_reflection_quat(who=supL, axis=i)
                    self.bo[supL, :] = quat_mult(q_supL,
                                                 self.bo[supL, :])
                elif self.bc[0] == 0:
                    self.pos[:, i] = torch.remainder(self.pos[:, i], self.L[i])
                else:
                    raise NotImplementedError()
            # Normalise for numerical stability
            self.bo = self.bo / torch.sqrt(
                torch.sum(self.bo ** 2, dim=1)).reshape((self.N, 1))
        else:
            raise NotImplementedError()



import math
import torch
from pykeops.torch import LazyTensor


def squared_distances_tensor(x, y, L, boundary_conditions='periodic'):
    """Squared distances matrix as torch Tensors.

    Args:
        x ((M,d) Tensor): Rows.
        y ((N,d) Tensor): Columns.
        L ((d,) Tensor): Box size.
        boundary_conditions (list or str, optional): Boundary conditions.
            Can be one of the following:

            - list of size :math:`d` containing 0 (periodic) or 1
              (wall with reflecting boundary conditions) for each
              dimension.

            - ``"open"`` : no boundary conditions.

            - ``"periodic"`` : periodic boundary conditions.

            - ``"spherical"`` : reflecting boundary conditions on the
              sphere of radius ``L[0]/2`` and center **L**/2.

            Default is ``"periodic"``.

    Returns:
        (M,N) Tensor:

        The tensor whose :math:`(i,j)` coordinate is the squared
        distance between ``x[i,:]`` and ``y[j,:]``.

    """
    d = x.size()[1]
    x_i = x[:, None, :]  # (M,1,d)
    y_j = y[None, :, :]  # (1,N,d)
    L_k = L[None, None, :]  # (1,1,d)
    A = y_j - x_i
    if boundary_conditions == 'periodic':
        A = A + ((-L_k / 2. - A) > 0) * L_k - ((A - L_k / 2.) > 0) * L_k
    elif (type(boundary_conditions) is list
          and not (boundary_conditions == [1] * d)):
        L_k = L_k * (-torch.tensor(boundary_conditions,
                                   dtype=L.dtype,
                                   device=L.device)[None, None, :] + 1.)
        A = A + ((-L_k / 2. - A) > 0) * L_k - ((A - L_k / 2.) > 0) * L_k
    squared_dist_ij = (A ** 2).sum(-1)
    return squared_dist_ij


def lazy_xy_matrix(x, y, L, boundary_conditions='periodic'):
    """XY matrix as LazyTensors.

    Args:
        x ((M,d) Tensor): Rows.
        y ((N,d) Tensor: Columns.
        L ((d,) Tensor): Box size.
        boundary_conditions (list or str, optional): Boundary conditions.
            Can be one of the following:

            - list of size :math:`d` containing 0 (periodic) or 1
              (wall with reflecting boundary conditions) for each
              dimension.

            - ``"open"`` : no boundary conditions.

            - ``"periodic"`` : periodic boundary conditions.

            - ``"spherical"`` : reflecting boundary conditions on the
              sphere of radius ``L[0]/2`` and center **L**/2.

            Default is ``"periodic"``.

    Returns:
        (M,N,d) LazyTensor:

        The LazyTensor ``A`` such that for each coordinate :math:`(i,j)`,
        ``A[i,j,:]`` is the LazyTensor ``LazyTensor(y[j,:]-x[i,:])``.

    """
    d = x.size()[1]
    x_i = LazyTensor(x[:, None, :])  # (M,1,d)
    y_j = LazyTensor(y[None, :, :])  # (1,N,d)
    L_k = LazyTensor(L[None, None, :])  # (1,1,d)
    A = y_j - x_i
    if boundary_conditions == 'periodic':
        A = A + (-L_k / 2. - A).step() * L_k - (A - L_k / 2.).step() * L_k
    elif (type(boundary_conditions) is list
          and not (boundary_conditions == [1] * d)):
        L_k = L_k * LazyTensor(-torch.tensor(boundary_conditions,
                                             dtype=L.dtype,
                                             device=L.device)[None, None, :]
                               + 1.)
        A = A + (-L_k / 2. - A).step() * L_k - (A - L_k / 2.).step() * L_k
    return A


def squared_distances(x, y, L, boundary_conditions='periodic'):
    """Squared distances LazyTensor.

    Args:
        x ((M,d) Tensor): Rows.
        y ((N,d) Tensor: Columns.
        L ((d,) Tensor): Box size.
        boundary_conditions (list or str, optional): Boundary conditions.
            Can be one of the following:

            - list of size :math:`d` containing 0 (periodic) or 1
              (wall with reflecting boundary conditions) for each
              dimension.

            - ``"open"`` : no boundary conditions.

            - ``"periodic"`` : periodic boundary conditions.

            - ``"spherical"`` : reflecting boundary conditions on the
              sphere of radius ``L[0]/2`` and center **L**/2.

            Default is ``"periodic"``.

    Returns:
        (M,N) LazyTensor:

        The LazyTensor whose :math:`(i,j)` coordinate is the squared
        distance between ``x[i,:]`` and ``y[j,:]``.

    """
    A = lazy_xy_matrix(x, y, L, boundary_conditions)
    squared_dist_ij = (A ** 2).sum(-1)
    return squared_dist_ij


def sqdist_angles(x, y, x_orientation, L, boundary_conditions='periodic'):
    """Squared distances and angles LazyTensors.

    Args:
        x ((M,d) Tensor): Rows.
        y ((N,d) Tensor: Columns.
        x_orientation ((M,d) Tensor): Orientation of x.
        L ((d,) Tensor): Box size.
        boundary_conditions (list or str, optional): Boundary conditions.
            Can be one of the following:

            - list of size :math:`d` containing 0 (periodic) or 1
              (wall with reflecting boundary conditions) for each
              dimension.

            - ``"open"`` : no boundary conditions.

            - ``"periodic"`` : periodic boundary conditions.

            - ``"spherical"`` : reflecting boundary conditions on the
              sphere of radius ``L[0]/2`` and center **L**/2.

            Default is ``"periodic"``.

    Returns:
        (M,N) LazyTensor, (M,N) LazyTensor:

        The LazyTensor whose :math:`(i,j)` coordinate is the squared
        distance between ``x[i,:]`` and ``y[j,:]`` and the LazyTensor
        whose :math:`(i,j)` coordinate is the cosine of the angle
        between ``y[j,:]-x[i,:]`` and ``x_orientation[i,:]``.

    """
    A = lazy_xy_matrix(x, y, L, boundary_conditions)
    squared_dist_ij = (A ** 2).sum(-1)
    cos_ij = (A * x_orientation[:, None, :]).sum(-1) / squared_dist_ij.sqrt()
    return squared_dist_ij, cos_ij


def lazy_interaction_kernel(x, y, Rx, L, boundary_conditions,
                            vision_angle=2 * math.pi, axis=None, **kwargs):
    """Interaction kernel as LazyTensor.

    Args:
        x ((M,d) Tensor): Row.
        y ((N,d) Tensor): Columns.
        Rx ((N,) Tensor or float): Interaction radius of x.
        L ((d,) Tensor): Box size.
        boundary_conditions (list or str, optional): Boundary conditions.
            Can be one of the following:

            - list of size :math:`d` containing 0 (periodic) or 1
              (wall with reflecting boundary conditions) for each
              dimension.

            - ``"open"`` : no boundary conditions.

            - ``"periodic"`` : periodic boundary conditions.

            - ``"spherical"`` : reflecting boundary conditions on the
              sphere of radius ``L[0]/2`` and center **L**/2.

            Default is ``"periodic"``.
        vision_angle (float, optional): Angle of vision of x. Default is
            :math:`2\pi`.
        axis ((M,d) Tensor, optional): Orientation of **x**. Default is
            ``None``. Must be specified if **vision_angle** is not
            :math:`2\pi`.
        **kwargs: Arbitrary keywords arguments.

    Returns:
        (M,N) LazyTensor:

        The LazyTensor whose :math:`(i,j)` coordinate is 1 if ``y[j,:]``
        is in the cone of vision of ``x[i,:]`` and 0 otherwise.

    """
    if type(Rx) == type(x):
        radius = LazyTensor(Rx.reshape((x.shape[0], 1))[:, None])
    else:
        radius = Rx
    if vision_angle == 2 * math.pi:
        lazy_squared_distances = squared_distances(x, y, L,
                                                   boundary_conditions)
        K_ij = (radius ** 2 - lazy_squared_distances).step()
    else:
        sqdist_ij, cos_ij = sqdist_angles(x, y, axis, L, boundary_conditions)
        K_ij = (radius ** 2 - sqdist_ij).step() \
               * (-float(math.cos(vision_angle / 2)) + cos_ij).step()
    return K_ij


def lazy_overlapping_kernel(x, y, Rx, Ry, L, boundary_conditions, **kwargs):
    """Overlapping kernel as LazyTensor.

    Args:
        x ((M,d) Tensor): Rows.
        y ((N,d) Tensor): Columns.
        Rx ((M,) Tensor or float): Radius of x.
        Ry ((N,) Tenosr or float): Radius of y.
        L ((d,) Tensor): Box size.
        boundary_conditions (list or str, optional): Boundary conditions.
            Can be one of the following:

            - list of size :math:`d` containing 0 (periodic) or 1
              (wall with reflecting boundary conditions) for each
              dimension.

            - ``"open"`` : no boundary conditions.

            - ``"periodic"`` : periodic boundary conditions.

            - ``"spherical"`` : reflecting boundary conditions on the
              sphere of radius ``L[0]/2`` and center **L**/2.

            Default is ``"periodic"``.
        **kwargs: Arbitrary keywords arguments.

    Returns:
        (M,N) LazyTensor:

        The LazyTensor whose :math:`(i,j)` coordinate is 1 if ``x[i,:]``
        and ``y[j,:]`` are at a distance smaller than ``Rx[i]+Ry[j]``
        (or smaller than ``Rx+Ry`` if all the **x** and **y** particles
        have the same radius).

    Raises:
        :exc:`NotImplementerError`: if **Rx** and **Ry** do not have the
             same type, or when **Rx** and **Ry** are float but **Rx**
             is not equal to **Ry**.

    """
    if not type(Rx) == type(Ry):
        raise NotImplementedError()
    if isinstance(Rx, float) and not (Rx == Ry):
        raise NotImplementedError()
    if type(Rx) == type(x):
        R_i = LazyTensor(Rx.reshape((x.shape[0], 1))[:, None])
        R_j = LazyTensor(Ry.reshape((y.shape[0], 1))[None, :])
        R_ij = R_i + R_j
        lazy_squared_distances = squared_distances(x, y, L,
                                                   boundary_conditions)
        K_ij = (R_ij ** 2 - lazy_squared_distances).step()
        return K_ij
    else:
        return lazy_interaction_kernel(x, y, 2 * Rx, L, boundary_conditions)


def lazy_morse(x, y, Ca, la, Cr, lr, p=2, mx=1., my=1., **kwargs):
    r"""LazyTensor of the forces derived from the Morse potential.

    The Morse potential is defined by

    .. math::
        U(z) = -C_a \exp(-|z|^p / \ell_a^p) + C_r \exp(-|z|^p / \ell_r^p)

    The force exerted by a particle located in :math:`y_j` on a particle
    located on :math:`x_i` is

    .. math::
        F = - m_x m_y  \nabla U(x_i - y_j)

    where :math:`m_x` and :math:`m_y` are the masses of the particles in
    :math:`x_i` and :math:`y_j` respectively.

    Args:
        x ((M,d) Tensor): Rows.
        y ((N,d) Tensor): Columns.
        Ca (float): Attraction coefficient.
        la (float): Attraction length.
        Cr (float): Repulsion coefficient.
        lr (float): Repulsion length.
        p (int, optional): Exponent. Default is 2.
        mx ((M,) Tensor or float, optional): Mass of **x**. Default is 1.
        my ((N,) Tensor or float, optional): Mass of **y**. Default is 1.
        **kwargs: Arbitrary keywords arguments.

    Returns:
        (M,N,d) LazyTensor:

        The LazyTensor whose :math:`(i,j,:)` coordinate is the force
        exerted by ``y[j,:]`` on ``x[i,:]`` derived from the Morse
        potential.

    """
    x_i = LazyTensor(x[:, None, :])  # (M,1,D)
    y_j = LazyTensor(y[None, :, :])  # (1,N,D)
    r = x_i - y_j
    sq_norm = (r ** 2).sum(-1)  # (M,N)
    if p == 2:
        attraction = Ca / (la ** p) * (-sq_norm / (la ** 2)).exp()
        repulsion = Cr / (lr ** p) * (-sq_norm / (lr ** 2)).exp()
        nablaU = 4 * (attraction - repulsion) * r
    else:
        sq_norm = sq_norm + (-sq_norm.sign() + 1.)  # case norm=0
        attraction = Ca / (la ** p) \
                     * (-(sq_norm ** (p / 2.)) / (la ** p)).exp()
        repulsion = Cr / (lr ** p) \
                    * (-(sq_norm ** (p / 2.)) / (lr ** p)).exp()
        nablaU = (p ** 2) * (attraction - repulsion) \
                 * (sq_norm ** ((p - 2.) / 2.)) * r
    if type(mx) == type(x):
        mx_i = LazyTensor(mx.reshape((x.shape[0], 1))[:, None])
        my_j = LazyTensor(my.reshape((y.shape[0], 1))[None, :])
        return - mx_i * my_j * nablaU
    else:
        return - mx * my * nablaU


def lazy_quadratic(x, y, R, L, boundary_conditions='periodic', **kwargs):
    r"""LazyTensor of the forces derived from of a quadratic potential.

    The quadratic potential is defined by:

    .. math::
        U(z) =  \frac{1}{2R} |z|^2 - |z|

    where :math:`R` is a fixed parameter. The force exerted on a
    particle located on :math:`y_j` on a particle located on :math:`x_i`
    is

    .. math::
        F = \left(\frac{1}{R}-\frac{1}{|y_j-x_i|}\right)(y_j-x_i).

    Args:
        x ((M,d) Tensor): Rows.
        y ((N,d) Tensor): Columns.
        R ((M,) Tensor or float): Radius of the **x** particles.
        L ((d,) Tensor): Box size.
        boundary_conditions (list or str, optional): Boundary conditions.
            Can be one of the following:

            - list of size :math:`d` containing 0 (periodic) or 1
              (wall with reflecting boundary conditions) for each
              dimension.

            - ``"open"`` : no boundary conditions.

            - ``"periodic"`` : periodic boundary conditions.

            - ``"spherical"`` : reflecting boundary conditions on the
              sphere of radius ``L[0]/2`` and center **L**/2.

            Default is ``"periodic"``.
        **kwargs: Arbitrary keywords arguments.

    Returns:
        (M,N,d) LazyTensor:

        The LazyTensor whose :math:`(i,j,:)` coordinate is the force
        exerted by ``y[j,:]`` on ``x[i,:]`` derived from the quadratic
        potential.

    """
    r = lazy_xy_matrix(x, y, L, boundary_conditions)
    sq_dist = (r ** 2).sum(-1)
    sq_dist = sq_dist + (-sq_dist.sign() + 1.)  # case norm=0
    dist = sq_dist.sqrt()
    if type(x) == type(R):
        Rx = LazyTensor(R.reshape((x.shape[0], 1))[:, None])
    else:
        Rx = R
    return (1. / Rx - 1. / dist) * r





import math
import torch
from .toolbox import angles_directions_to_quat, quat_mult


def uniform_sphere_rand(N, d, device, dtype=torch.float32):
    """Uniform sample on the unit sphere :math:`\mathbb{S}^{d-1}`.

    Args:
        N (int): Number of samples.
        d (int): Dimension
        device (torch.device): Device on which the samples are created
        dtype (torch.dtype, optional): Default is torch.float32

    Returns:
        (N,D) Tensor:

        Sample.

    """
    V = torch.randn(N, d, device=device, dtype=dtype)
    V = V / torch.norm(V, dim=1).reshape(N, 1)
    if (torch.isnan(V)).sum() > 0:
        return uniform_sphere_rand(N, d, device, dtype=dtype)
    else:
        return V


def uniform_angle_rand(mu, eta):
    """Uniform sample in an angle interval in dimension 2.

    Add a uniform angle in :math:`[-\pi\eta,\pi\eta]` to the angle of
    **mu**.

    Args:
        mu ((N,2) Tensor): Center of the distribution for each of the
            N samples.
        eta ((N,2) Tensor or float): Deviation for each of the N samples.

    Returns:
        (N,2) Tensor:

        Sample.

    """
    N, D = list(mu.size())
    angle_mu = torch.atan2(mu[:, 1], mu[:, 0])
    u = 2 * math.pi * eta * torch.rand(N,
                                       dtype=mu.dtype,
                                       device=mu.device) - math.pi * eta
    sample_angle = angle_mu + u
    result = torch.cat(
        (torch.cos(sample_angle).reshape(N, 1),
         torch.sin(sample_angle).reshape(N, 1)),
        dim=1)
    return result


""" Rejection sampler for the von Mises Fisher distribution 
on the (D-1)-sphere using the Ulrich-Wood algorithm
"""

def sample_W(kappa,N,d):
    """Sample the first coordinate of a von Mises distribution with
    center at :math:`(1,0,...0)^T` and concentration **kappa**.

    Args:
        kappa ((N,) Tensor): Concentration parameter.
            Also accept (1,) Tensor if all the samples have the same
            concentration parameter.
        N (int): Number of samples. Must be equal to the size of kappa
            when the size of kappa is >1.
        d: Dimension.

    Returns:
        (N,) Tensor:

        Random sample in :math:`\mathbb{R}`.

    """
    if N == 0:
        return torch.tensor([],device=kappa.device)
    b = (-2 * kappa + torch.sqrt(4 * (kappa ** 2) + (d - 1) ** 2)) / (d - 1)
    x0 = (1-b) / (1+b)
    c = kappa * x0 + (d - 1) * torch.log(1 - x0 ** 2)
    X = kappa.new(2 * N, d - 1).normal_()
    X = (X**2).sum(1)
    X0 = X[:N]
    X1 = X[N:]
    beta = X0 / (X0 + X1)   #beta is a Beta((D-1)/2,(D-1)/2)
    w = (1 - (1+b) * beta) / (1 - (1-b) * beta)
    W = w
    U = kappa.new(int(N)).uniform_()
    reject = kappa * w + (d - 1) * torch.log(1 - x0 * w) - c < torch.log(U)
    Nreject = reject.sum().item()
    if kappa.shape[0] == 1:
        W[reject] = sample_W(kappa, Nreject, d)
    else:
        W[reject] = sample_W(kappa[reject], Nreject, d)
    return W


def vonmises_rand(mu,kappa):
    """Von Mises sample on the sphere :math:`\mathbb{S}^{d-1}` using
    the Ulrich-Wood algorithm [W1994]_.

    .. [W1994] A. Wood, Simulation of the von Mises Fisher distribution,
       *Comm. Statist. Simulation Comput.*, Vol. 23, No. 1 (1994).

    Args:
        mu ((N,d) Tensor): Centers of the N samples.
        kappa: ((N,) Tensor or (1,) Tensor): Concentration parameter.

    Returns:
        (N,d) Tensor:

        Sample with center **mu** and concentration parameter **kappa**.

    """
    N, D = list(mu.size())
    W = sample_W(kappa, N, D)
    # Uniformly sampled unit random vector V in the (D-2)-sphere
    if D == 2:
        u = torch.rand((N, 1), dtype=mu.dtype, device=mu.device) < .5
        V = 2. * u - 1.
    else:
        V = torch.randn((N, D - 1), dtype=mu.dtype, device=mu.device)
        V = V / torch.norm(V, dim=1).reshape((N, 1))
    # Construct a von Mises random variable X with center (1,0,0...)^T
    X = torch.cat((W[:, None], torch.sqrt((1 - W[:, None] ** 2)) * V), 1)
    if D == 2:
        angle_mu = torch.atan2(mu[:, 1], mu[:, 0])
        X_angle = torch.atan2(X[:, 1], X[:, 0])
        X_angle = X_angle + angle_mu
        result = torch.cat(
            (torch.cos(X_angle).reshape(N, 1),
             torch.sin(X_angle).reshape(N, 1)),
            dim=1)
    else:
        # Construct the Householder matrix Q associated to mu
        e1 = torch.zeros(D, dtype=mu.dtype, device=mu.device)
        e1[0] = 1
        u = mu - e1[None, :]
        v = u / torch.norm(u, dim=1).reshape((N, 1))
        v[torch.isnan(v)] = 0
        Q = torch.eye(D, dtype=mu.dtype, device=mu.device) - \
            2 * torch.einsum('ijl,ilk->ijk', v[:, :, None], v[:, None, :])
        # Multiply the Householder matrix with X to get the correct result
        result = torch.einsum('ijl,ilk->ijk', Q, X[:, :, None])
        result = result.reshape(N, D)
    return result

#### Rejection sampler: uniform quaternion in a ball around Id ####

def sample_angles(N, scales):
    """Sample angles by rejection sampling."""
    # if scales.size()==torch.Size([1]):
    #     s = scales
    # else:
    #     s = scales.reshape(N,1)
    angles = scales * torch.rand(N, dtype=scales.dtype, device=scales.device)
    uniform = torch.rand(N, dtype=scales.dtype, device=scales.device)

    threshold = (1 - angles.cos()) / (1 - scales.cos())
    reject = (uniform > threshold).view(-1)

    M = int(reject.sum())

    if M == 0:
        return angles
    else:
        if scales.size() == torch.Size([1]):
            angles[reject] = sample_angles(M, scales)
        else:
            angles[reject] = sample_angles(M, scales[reject])
        return angles


def uniformball_unitquat_rand(q, scales):
    """Sample rotations at random in balls centered around q,
    with radii given by the scales array."""

    N, d = list(q.size())
    angles = sample_angles(N, scales).reshape(N, 1)
    # Direction, randomly sampled on the sphere
    directions = torch.randn(N, 3, dtype=scales.dtype, device=scales.device)
    directions = directions / torch.norm(directions, dim=1).reshape((N, 1))

    q0 = angles_directions_to_quat(angles, directions)  # Uniform around 1

    return quat_mult(q, q0)


def uniform_unitquat_rand(N, device, dtype=torch.float32):
    """Uniform sample in the space of unit quaternions.

    Args:
        N (int): Number of samples.
        device (torch.device): Device on which the samples are created.
        dtype (torch.dtype, optional): Default is torch.float32.

    Returns:
        (N,4) Tensor:

        Samples.

    """
    angles = sample_angles(N, torch.tensor([math.pi],
                                           dtype=dtype, device=device))
    directions = torch.randn(N, 3, dtype=dtype, device=device)
    directions = directions / torch.norm(directions, dim=1).reshape((N, 1))
    return angles_directions_to_quat(angles, directions)


""" Rejection sampler for the von Mises Fisher distribution 
on the space of unit quaternions using the BACG method.
"""

def sample_q0(kappa, N):
    """Sample **N** random variables distributed according to the von
    Mises distribution on the space of quaternions with center
    :math:`(1,0,0,0)` and concentration parameter **kappa**.

    Use the BACG method described in [KGM2018]_.

    .. [KGM2018] J. Kent, A. Ganeiber & K. Mardia, A New Unified Approach
       for the Simulation of a Wide Class of Directional Distributions,
       *J. Comput. Graph. Statist.*, Vol. 27, No. 2 (2018).

    Args:
        kappa ((N,) Tensor or (1,) Tensor): Concentration parameter.
        N (int): Number of samples. Must be equal to the size of kappa
            when the size of kappa is >1.

    Returns:
        (N,) Tensor:

        Sample.

    """
    if N == 0:
        return torch.tensor([], device=kappa.device)
    ell = 2 * kappa
    b0 = torch.sqrt((ell - 2) ** 2 + 2 * ell) - (ell - 2)
    logM = -.5 * (4 - b0) + 2 * torch.log(4 / b0)
    if kappa.shape[0] == 1:
        A = torch.tensor([0., 2. * kappa, 2. * kappa, 2. * kappa],
                         dtype=kappa.dtype, device=kappa.device)
        Omega = torch.ones(4,
                           dtype=kappa.dtype,
                           device=kappa.device) + (2 / b0) * A
        Sigma = Omega ** (-1)
    else:
        A = torch.cat((
            torch.zeros(N, 1, dtype=kappa.dtype, device=kappa.device),
            (2. * kappa).reshape(N, 1),
            (2. * kappa).reshape(N, 1),
            (2. * kappa).reshape(N, 1)),
            dim=1)
        Omega = torch.ones(1, 4, dtype=kappa.dtype, device=kappa.device) \
                + (2. / b0.reshape(N, 1)) * A
        Sigma = Omega ** (-1)
    y = torch.sqrt(Sigma) * torch.randn(N, 4,
                                        dtype=kappa.dtype, device=kappa.device)
    q0 = y / torch.norm(y, dim=1).reshape((N, 1))
    U = torch.rand(N, dtype=kappa.dtype, device=kappa.device)
    u = torch.sum(A * q0 ** 2, dim=1)
    log_acg = -2 * torch.log(torch.sum(Omega * q0 ** 2, dim=1))
    reject = torch.log(U) > -u - logM - log_acg
    Nreject = reject.sum()
    if Nreject == 0:
        return q0
    else:
        if kappa.shape[0] == 1:
            q0[reject, :] = sample_q0(kappa, Nreject)
        else:
            q0[reject, :] = sample_q0(kappa[reject], Nreject)
        return q0


def vonmises_quat_rand(q, kappa):
    """ Von Mises random variables on the space of quaternions with
    center **q** and concentration parameter **kappa**.

    Multiply **q** with a sample from the von Mises distribution on the
    space of quaternions with center :math:`(1,0,0,0)` and concentration
    parameter **kappa**. See :func:`sample_q0`.

    Args:
        q ((N,4) Tensor): Center of the distribution.
        kappa ((N,) Tensor or (1,) Tensor): Concentration parameter.

    Returns:
        (N,4) Tensor:

        Sample.

    """
    q0 = sample_q0(kappa, q.shape[0])
    return quat_mult(q, q0)




import math
import torch
import numpy as np
import matplotlib as mpl
from matplotlib import pyplot as plt
from matplotlib.collections import EllipseCollection
import sys
import os
import pickle


def save(simu, frames, attributes, attributes_alltimes, Nsaved=None,
         save_file=True, file_name="data"):
    """Basic saving function.

    Args:
        simu (Particles): A model.
        frames (list): List of times (float) to save.
        attributes (list): List of strings containing the names of the
            attributes to save at the times in the list **frames**.
        attributes_alltimes (list): List of strings containing the
            names of the attributes which will be saved at each
            iteration until the last time in frame.
        Nsaved (int, optional): Default is None which means that all
            the particles will be saved.
        save_file: (bool, optional): Default is True which means that
            a pickle file will be generated and saved in the current
            directory.
        file_name (str, optional): Default is ``"data"``.


    Returns:
        dict:

        A dictionary which contains the entry keywords ``"time"``,
        ``"frames"`` and all the elements of the lists **attributes**
        and **attributes_alltimes**. The dictionary is saved in the
        current directory.

    """
    frames = sorted(frames)
    if Nsaved is None:
        N = simu.N
    else:
        N = Nsaved
    data = {"time" : [0],
            "frames" : [0]}
    for att in attributes_alltimes:
        if type(att) is str:
            x = getattr(simu, att)
            att_name = att
        else:
            x = att(simu)
            att_name = att.__name__
        if torch.is_tensor(x):
            if x.shape[0] == simu.N:
                x = x[:N].clone()
        data[att_name] = [x]
    for att in attributes:
        if type(att) is str:
            x = getattr(simu, att)
            att_name = att
        else:
            x = att(simu)
            att_name = att.__name__
        if torch.is_tensor(x):
            if x.shape[0] == simu.N:
                x = x[:N].clone()
        data[att_name] = [x]
    t = 0
    tmax = len(frames) - 1
    percent = 0
    while t < tmax+1:
        simu.__next__()
        data["time"].append(data["time"][-1] + simu.dt)
        if data["time"][-1] / frames[tmax] >= percent / 100.:
            sys.stdout.write('\r' + "Progress:" + str(percent) + "%")
            sys.stdout.flush()
            percent += 1
        if Nsaved is None:
            N = simu.N
        else:
            N = Nsaved
        for att in attributes_alltimes:
            if type(att) is str:
                x = getattr(simu, att)
                att_name = att
            else:
                x = att(simu)
                att_name = att.__name__
            if torch.is_tensor(x):
                if x.shape[0] == simu.N:
                    x = x[:N].clone()
            data[att_name].append(x)
        if abs(data["time"][-1] >= frames[t]):
            data["frames"].append(data["time"][-1])
            for att in attributes:
                if type(att) is str:
                    x = getattr(simu, att)
                    att_name = att
                else:
                    x = att(simu)
                    att_name = att.__name__
                if torch.is_tensor(x):
                    if x.shape[0] == simu.N:
                        x = x[:N].clone()
                data[att_name].append(x)
            t += 1
    if save_file:
        pickle.dump(data, open(file_name + ".p", "wb"))
    return data


def display(position, velocity):
    # Quiver plot in dimension 2
    cpu_pos = position.cpu()
    cpu_vel = velocity.cpu()
    U = cpu_vel[:, 0]
    V = cpu_vel[:, 1]
    plot = plt.quiver(cpu_pos[:, 0], cpu_pos[:, 1], U, V)
    return plot


def display_kinetic_particles(simu, time, N_dispmax=None,
                              order=False, color=False,
                              show=True, figsize=(6,6),
                              save=False, path='simu'):
    """Basic quiver plot in dimension 2 and scatter plot in dimension 3.

    Args:
        simu (Particles): A model.
        time (list): List of times to plot.
        N_dispmax (int, optional): Default is None.
        order (bool, optional): Compute the order parameter or not.
            Default is False.
        color (bool, optional): The color of the particle depends on
            the velocity angle (dimension 2). Default is False.
        show (bool, optional): Show the plot. Default is True.
        figsize (tuple, optional): Figure size, default is (6,6).
        save (bool, optional): Save the plots. Default is False.
        path (str, optional): The plots will be saved in the directory
            ``'./path/frames'``.

    Returns:
        list, list:

        Times and order parameter

    """
    t = 0
    tmax = len(time) - 1
    percent = 0
    op = []
    real_time = []
    if save:
        current_directory = os.getcwd()
        final_directory = os.path.join(current_directory, path, r'frames')
        if not os.path.exists(final_directory):
            os.makedirs(final_directory)
    for it, info in enumerate(simu):
        if N_dispmax is None:
            N_displayed = simu.N
        else:
            N_displayed = min(N_dispmax, simu.N)
        real_time.append(it * simu.dt)
        if order:
            op.append(simu.order_parameter)
        if it == np.floor((percent / 100) * time[tmax] / simu.dt):
            sys.stdout.write('\r' + "Progress:" + str(percent) + "%")
            sys.stdout.flush()
            percent += 1
        if abs(it * simu.dt - time[t]) < simu.dt:
            if simu.d == 2:
                f = plt.figure(t, figsize=figsize)
                ax = f.add_subplot(111)
                x = simu.pos[:N_displayed, 0]
                y = simu.pos[:N_displayed, 1]
                U = simu.vel[:N_displayed, 0]
                V = simu.vel[:N_displayed, 1]
                if color:
                    plt.set_cmap('hsv')
                    angle = torch.atan2(V, U)
                    quiv = ax.quiver(x.cpu(), y.cpu(), U.cpu(), V.cpu(),
                                     angle.cpu())
                    cb = ax.figure.colorbar(quiv, ax=ax,
                                            ticks=[-3.1415, 0, 3.1415],
                                            fraction=0.046, pad=0.04)
                else:
                    ax.quiver(x.cpu().numpy(), y.cpu().numpy(),
                              U.cpu().numpy(), V.cpu().numpy())
                if simu.bc == 'spherical':
                    theta = np.linspace(0, 2 * math.pi, 1000)
                    r = simu.L[0].cpu() / 2
                    ax.plot(r + r * np.cos(theta), r + r * np.sin(theta))
                ax.set_xlim(xmin=0, xmax=simu.L[0].cpu())
                ax.set_ylim(ymin=0, ymax=simu.L[1].cpu())
                ax.set_aspect('equal')
                ax.set_title(
                    simu.name + '\n Parameters: ' + simu.parameters
                    + '\n Time=' + str(round(simu.iteration * simu.dt, 1)),
                    fontsize=10)
            elif simu.d == 3:
                f = plt.figure(t, figsize=figsize)
                ax = f.add_subplot(111, projection='3d')
                X = simu.pos[:N_displayed, 0].cpu()
                Y = simu.pos[:N_displayed, 1].cpu()
                Z = simu.pos[:N_displayed, 2].cpu()
                ax.scatter(X, Y, Z, s=0.1)
                ax.set_xlim(xmin=0, xmax=simu.L[0].cpu())
                ax.set_ylim(ymin=0, ymax=simu.L[1].cpu())
                ax.set_zlim(zmin=0, zmax=simu.L[2].cpu())
                ax.set_title(
                    simu.name + '\n Parameters: ' + simu.parameters
                    + '\n Time=' + str(round(simu.iteration * simu.dt, 1)),
                    fontsize=10)
            if save:
                f.savefig(f"{final_directory}/" + str(t) + ".png")
            if show:
                # plt.show()
                plt.pause(0.000001)
            else:
                plt.close()
            t += 1
        if t > tmax:
            break

    if order:
        x = np.array(real_time)
        y = np.array(op)
        op_plot = plt.figure(t, figsize=(6, 6))
        plt.plot(x, y)
        plt.xlabel('time')
        plt.ylabel('order parameter')
        plt.axis([0, time[tmax], 0, 1])
        if save:
            simu_dir = os.path.join(current_directory, path)
            op_plot.savefig(f"{simu_dir}/order_parameter.png")
        return x, y
    else:
        return None, None


def scatter_particles(simu, time, show=True,
                      save=False, path='simu'):
    """Scatter plot with the radii of the particles.

    Args:
        simu (Particles): A model.
        time (list): List of times to plot.
        show (bool, optional): Show the plot. Default is True.
        save (bool, optional): Save the plots. Default is False.
        path (str, optional): The plots will be saved in the directory
            ``'./path/frames'``.

    """
    # Scatter plots at times specified by the list 'time' for the
    # mechanism 'mechanism' (iterator inherited from the class 'particle')
    t = 0
    tmax = len(time) - 1
    percent = 0
    real_time = [0.]
    if save:
        current_directory = os.getcwd()
        final_directory = os.path.join(current_directory, path, r'frames')
        if not os.path.exists(final_directory):
            os.makedirs(final_directory)
    for it, info in enumerate(simu):
        real_time.append(real_time[-1] + simu.dt)
        if real_time[-1] / time[tmax] >= percent / 100:
            sys.stdout.write('\r' + "Progress:" + str(percent) + "%")
            sys.stdout.flush()
            percent += 1
        if abs(real_time[-1] >= time[t]):
            if simu.d == 2:
                f = plt.figure(t, figsize=(6, 6))
                ax = f.add_subplot(111)
                x = simu.pos[:, 0].cpu()
                y = simu.pos[:, 1].cpu()
                size = 2 * simu.R.cpu()
                offsets = list(zip(x, y))
                ax.add_collection(EllipseCollection(
                    widths=size, heights=size, angles=0, units='xy',
                    edgecolor='k', offsets=offsets, transOffset=ax.transData))
                ax.axis('equal')  # set aspect ratio to equal
                ax.axis([0, simu.L[0].cpu(), 0, simu.L[1].cpu()])
                #                 ax.set_aspect(1)
                #                 plt.axes().set_aspect('equal', 'box')
                ax.set_title(simu.name + '\n Parameters: ' + simu.parameters
                             + '\n Time=' + str(round(real_time[-1], 1)),
                             fontsize=10)
            elif simu.d == 3:
                f = plt.figure(t, figsize=(10, 10))
                ax = f.add_subplot(111, projection='3d')
                X = simu.pos[:, 0].cpu()
                Y = simu.pos[:, 1].cpu()
                Z = simu.pos[:, 2].cpu()
                ax.scatter(X, Y, Z, s=0.1)
                ax.set_xlim(0, simu.L[0].cpu())
                ax.set_ylim(0, simu.L[1].cpu())
                ax.set_zlim(0, simu.L[2].cpu())
                ax.set_title(
                    simu.name + '\n Parameters: ' + simu.parameters
                    + '\n Time=' + str(round(simu.iteration * simu.dt, 1)),
                    fontsize=10)
            if save:
                f.savefig(f"{final_directory}/" + str(t) + ".png")
            if show:
                # plt.show()
                plt.pause(0.000001)
            else:
                plt.close()
            t += 1
        if t > tmax:
            break


def heatmap_particles(mechanism, time, N_displayed, nbins, max_density=25, order=False, show=True, save=False,
                      path='simu'):
    # Scatter plots at times specified by the list 'time' for the
    # mechanism 'mechanism' (iterator inherited from the class 'particle')
    t = 0
    tmax = len(time) - 1
    percent = 0
    real_time = [0.]
    op = [mechanism.order_parameter]
    if save:
        current_directory = os.getcwd()
        final_directory = os.path.join(current_directory, path, r'frames')
        if not os.path.exists(final_directory):
            os.makedirs(final_directory)
    for it, info in enumerate(mechanism):
        real_time.append(real_time[-1] + mechanism.dt)
        if order:
            op.append(mechanism.order_parameter)
        if real_time[-1] / time[tmax] >= percent / 100:
            sys.stdout.write('\r' + "Progress:" + str(percent) + "%")
            sys.stdout.flush()
            percent += 1
        if abs(real_time[-1] >= time[t]):
            f = plt.figure(t, figsize=(12, 6))
            ax = f.add_subplot(121)

            x = mechanism.pos[0:N_displayed, 0].cpu().numpy()
            y = mechanism.pos[0:N_displayed, 1].cpu().numpy()
            L0 = float(mechanism.L[0].cpu())
            L1 = float(mechanism.L[1].cpu())
            counts, xedges, yedges, im = ax.hist2d(x, y, bins=nbins, range=np.array([[0, L0], [0, L0]]), density=True)
            ax.axis('equal')  # set aspect ratio to equal
            ax.axis([0, L0, 0, L1])
            #                 ax.set_aspect(1)
            #                 plt.axes().set_aspect('equal', 'box')
            ax.set_title(mechanism.name + '\n Parameters: ' + mechanism.parameters
                         + '\n Time=' + str(round(real_time[-1], 1)), fontsize=10)
            im.set_clim(0, max_density)
            cb = f.colorbar(im, ax=ax)

            ax1 = f.add_subplot(122)
            #             U = mechanism.vel[0:N_displayed,0].cpu()
            #             V = mechanism.vel[0:N_displayed,1].cpu()
            #             ax1.quiver(x,y,U,V)
            #             ax1.axis('equal')
            #             ax1.axis([0,L0,0,L1])

            if save:
                f.savefig(f"{final_directory}/" + str(t) + ".png")
            if show:
                plt.show()
            else:
                plt.close()
            t += 1
        if t > tmax:
            break
    if order:
        x = np.array(real_time)
        y = np.array(op)
        op_plot = plt.figure(t, figsize=(6, 6))
        plt.plot(x, y)
        plt.xlabel('time')
        plt.ylabel('order parameter')
        plt.axis([0, time[tmax], 0, 1])
        if save:
            simu_dir = os.path.join(current_directory, path)
            op_plot.savefig(f"{simu_dir}/order_parameter.png")
        return x, y


def scatter_particles_Poliakov(mechanism, time, N_displayed, show=True, save=False, path='simu'):
    # Scatter plots at times specified by the list 'time' for the
    # mechanism 'mechanism' (iterator inherited from the class 'particle')
    t = 0
    tmax = len(time) - 1
    percent = 0
    real_time = [0.]
    if save:
        current_directory = os.getcwd()
        final_directory = os.path.join(current_directory, path, r'frames')
        if not os.path.exists(final_directory):
            os.makedirs(final_directory)
    for it, info in enumerate(mechanism):
        real_time.append(real_time[-1] + mechanism.dt)
        if real_time[-1] / time[tmax] >= percent / 100:
            sys.stdout.write('\r' + "Progress:" + str(percent) + "%")
            sys.stdout.flush()
            percent += 1
        if abs(real_time[-1] >= time[t]):
            f = plt.figure(t, figsize=(6, 6))
            ax = f.add_subplot(111)

            xA = mechanism.pos[mechanism.A, 0].cpu()
            yA = mechanism.pos[mechanism.A, 1].cpu()
            xAe = xA[mechanism.sigma == 1]
            yAe = yA[mechanism.sigma == 1]
            xAs = xA[mechanism.sigma == 0]
            yAs = yA[mechanism.sigma == 0]
            xB = mechanism.pos[mechanism.B, 0].cpu()
            yB = mechanism.pos[mechanism.B, 1].cpu()
            size = 2 * mechanism.lr

            edgecolor = None

            offsets = list(zip(xB, yB))
            ax.add_collection(EllipseCollection(widths=size, heights=size, angles=0, units='xy',
                                                edgecolor=edgecolor, facecolor='k', offsets=offsets,
                                                transOffset=ax.transData))

            offsets = list(zip(xAe, yAe))
            ax.add_collection(EllipseCollection(widths=size, heights=size, angles=0, units='xy',
                                                edgecolor=edgecolor, facecolor='r', offsets=offsets,
                                                transOffset=ax.transData))

            offsets = list(zip(xAs, yAs))
            ax.add_collection(EllipseCollection(widths=size, heights=size, angles=0, units='xy',
                                                edgecolor=edgecolor, facecolor='b', offsets=offsets,
                                                transOffset=ax.transData))

            ax.axis('equal')  # set aspect ratio to equal
            ax.axis([0, mechanism.L[0].cpu(), 0, mechanism.L[1].cpu()])
            #                 ax.set_aspect(1)
            #                 plt.axes().set_aspect('equal', 'box')
            ax.set_title(mechanism.name + '\n Parameters: ' + mechanism.parameters
                         + '\n Time=' + str(round(real_time[-1], 1)), fontsize=10)
            if save:
                f.savefig(f"{final_directory}/" + str(t) + ".png")
            if show:
                plt.show()
            else:
                plt.close()
            t += 1
        if t > tmax:
            break


def scatter_particles_Walczak(mechanism, time, N_displayed, order=True, show=True, save=False, path='simu'):
    # Scatter plots at times specified by the list 'time' for the
    # mechanism 'mechanism' (iterator inherited from the class 'particle')
    t = 0
    tmax = len(time) - 1
    percent = 0
    real_time = [0.]
    op = [mechanism.order_parameter]
    if save:
        current_directory = os.getcwd()
        final_directory = os.path.join(current_directory, path, r'frames')
        if not os.path.exists(final_directory):
            os.makedirs(final_directory)
    for it, info in enumerate(mechanism):
        real_time.append(real_time[-1] + mechanism.dt)
        if order:
            op.append(mechanism.order_parameter)
        if real_time[-1] / time[tmax] >= percent / 100:
            sys.stdout.write('\r' + "Progress:" + str(percent) + "%")
            sys.stdout.flush()
            percent += 1
        if abs(real_time[-1] >= time[t]):
            f = plt.figure(t, figsize=(6, 6))
            ax = f.add_subplot(111)

            x = mechanism.pos[:, 0].cpu()
            y = mechanism.pos[:, 1].cpu()
            phi = mechanism.phase.cpu().numpy() / (2 * math.pi)
            size = 2 * mechanism.R

            edgecolor = None
            color = phi

            ax.scatter(x, y, c=color, cmap='hsv')

            ax.axis('equal')  # set aspect ratio to equal
            ax.axis([0, mechanism.L[0].cpu(), 0, mechanism.L[1].cpu()])
            ax.set_title(mechanism.name + '\n Parameters: ' + mechanism.parameters
                         + '\n Time=' + str(round(real_time[-1], 1)), fontsize=10)
            if save:
                f.savefig(f"{final_directory}/" + str(t) + ".png")
            if show:
                plt.show()
            else:
                plt.close()
            t += 1
        if t > tmax:
            break
    if order:
        x = np.array(real_time)
        y = np.array(op)
        op_plot = plt.figure(t, figsize=(6, 6))
        plt.plot(x, y)
        plt.xlabel('time')
        plt.ylabel('order parameter')
        plt.axis([0, time[tmax], 0, 1])
        if save:
            simu_dir = os.path.join(current_directory, path)
            op_plot.savefig(f"{simu_dir}/order_parameter.png")
        return x, y


def save_bo(mechanism, to_save, Nmax=2000, order=False, path='data'):
    N_saved = min(Nmax, mechanism.N)
    percent = 0
    theta = []
    phi = []
    real_time = []
    should_save = False
    savei = 0
    max_time = to_save[-1]
    current_directory = os.getcwd()
    final_directory = os.path.join(current_directory, r"" + path)
    if not os.path.exists(final_directory):
        os.makedirs(final_directory)
    for it, info in enumerate(mechanism):
        real_time.append(it * mechanism.dt)
        theta.append(mechanism.theta)
        phi.append(mechanism.phi)

        if it == np.floor((percent / 100) * max_time / mechanism.dt):
            sys.stdout.write('\r' + "Progress:" + str(percent) + "%")
            sys.stdout.flush()
            percent += 1
        if it * mechanism.dt > max_time:
            break

        should_save = it * mechanism.dt >= to_save[savei]
        if should_save:
            torch.save(mechanism.pos[range(N_saved), :], final_directory + '/pos' + str(savei) + '.pt')
            torch.save(mechanism.bo[range(N_saved), :], final_directory + '/bo' + str(savei) + '.pt')
            savei += 1
    x = np.array(real_time)
    y = np.array(theta)
    z = np.array(phi)
    theta_plot = plt.figure(1, figsize=(6, 6))
    plt.plot(x, y)
    plt.xlabel('time')
    plt.ylabel('theta')
    plt.axis([0, max_time, 0, 3.15])
    theta_plot.savefig(f"{final_directory}/theta.png")
    phi_plot = plt.figure(2, figsize=(6, 6))
    plt.plot(x, z)
    plt.xlabel('time')
    plt.ylabel('phi')
    plt.axis([0, max_time, 0, 6.3])
    theta_plot.savefig(f"{final_directory}/phi.png")
    param = open(final_directory + '/parameters.txt', 'w')
    param.write(mechanism.parameters)
    param.close()
    times = open(final_directory + '/times_saved.txt', 'w')
    times.write(str(to_save))
    times.close()
    np.save(final_directory + '/times.npy', x)
    np.save(final_directory + '/theta.npy', y)
    np.save(final_directory + '/phi.npy', z)
    return real_time, theta, phi


def plot_bo_twist_order(mechanism,
                        to_plot,
                        K,
                        axis,
                        Nmax=2000,
                        show=True,
                        save_particles=False,
                        save_twist=False,
                        save_order=False,
                        path='data'):
    N_saved = min(Nmax, mechanism.N)
    percent = 0
    theta = []
    phi = []
    real_time = []
    local_order = []
    global_order = []
    ordertime = []
    should_plot = False
    ploti = 0
    which_axis = len(axis)
    colors = range(K)
    max_time = to_plot[-1]
    slices = torch.zeros((K, 2, len(to_plot), len(axis))).type_as(mechanism.pos)
    if save_particles or save_twist or save_order:
        current_directory = os.getcwd()
        final_directory = os.path.join(current_directory, r"" + path)
        if not os.path.exists(final_directory):
            os.makedirs(final_directory)
        if save_particles:
            os.makedirs(os.path.join(final_directory, r'particles'))
        if save_twist:
            os.makedirs(os.path.join(final_directory, r'twist'))
    for it, info in enumerate(mechanism):
        real_time.append(it * mechanism.dt)
        theta.append(mechanism.theta)
        phi.append(mechanism.phi)

        if it == np.floor((percent / 100) * max_time / mechanism.dt):
            sys.stdout.write('\r' + "Progress:" + str(percent) + "%")
            sys.stdout.flush()
            percent += 1
        if it * mechanism.dt > max_time:
            break

        should_plot = it * mechanism.dt >= to_plot[ploti]
        if should_plot:
            if save_particles:
                torch.save(mechanism.pos[range(N_saved), :], final_directory + '/particles/pos' + str(ploti) + '.pt')
                torch.save(mechanism.bo[range(N_saved), :], final_directory + '/particles/bo' + str(ploti) + '.pt')
            if save_order:
                ordertime.append(it * mechanism.dt)
                local_order.append(mechanism.local_order)
                global_order.append(mechanism.global_order)
            for ax in range(which_axis):
                slice_p, slice_q = mechanism.twist(K, axis[ax])
                slices[:, :, ploti, ax] = torch.cat((slice_p.reshape((K, 1)), slice_q.reshape((K, 1))), axis=1)
                slice_p = slice_p.cpu()
                slice_q = slice_q.cpu()
                axf, axi = plt.subplots(figsize=(10, 9))
                angle = np.arange(0., 2 * np.pi, 0.01)
                axi.plot(np.cos(angle), np.sin(angle), linewidth=6)
                center = axi.scatter([0.], [0.], s=200, marker='X')
                scat = axi.scatter(slice_p, slice_q, c=colors, cmap='hsv')

                axi.set_xlim(-1.1, 1.1)
                axi.set_xticks([-1, 0, 1])
                axi.set_xticklabels([-1, 0, 1], fontsize=48)
                axi.set_ylim(-1.1, 1.1)
                axi.set_yticks([-1, 0, 1])
                axi.set_yticklabels([-1, 0, 1], fontsize=48)
                axi.set_aspect('equal', 'box')
                axi.set_title('Twist along the axis ' + axis[ax] + ' \n Time=' + str(round(to_plot[ploti], 1)),
                              fontsize=36)
                cb = axi.figure.colorbar(scat, ax=axi, fraction=0.046, pad=0.04)
                cb.ax.tick_params(labelsize=36)
                if show:
                    plt.show()
                else:
                    plt.close()
                if save_twist:
                    axf.savefig(f"{final_directory}/twist/twist_" + axis[ax] + "_" + str(ploti) + ".png")
            ploti += 1
    if save_twist:
        torch.save(slices, final_directory + '/slices.pt')
    x = np.array(real_time)
    y = np.array(theta)
    z = np.array(phi)
    theta_plot = plt.figure(1, figsize=(12, 12))
    plt.plot(x, y, linewidth=12)
    plt.xlabel('time', fontsize=36)
    plt.xticks(ticks=[0, max_time / 4, max_time / 2, 3 * max_time / 4, round(max_time)],
               labels=[0, round(max_time / 4), round(max_time / 2), round(3 * max_time / 4), max_time], fontsize=36)
    plt.ylabel('θ', fontsize=36, rotation=0, labelpad=20)
    plt.yticks(fontsize=36)
    plt.yticks(ticks=[0, np.pi / 2, np.pi], labels=['0', 'π/2', 'π'], fontsize=36)
    plt.hlines([0, np.pi / 2, np.pi], xmin=0, xmax=max_time, linestyles='dashed', linewidth=4)
    plt.axis([0, max_time, 0, 3.15])

    phi_plot = plt.figure(2, figsize=(12, 12))
    plt.plot(x, z, linewidth=12)
    plt.xlabel('time', fontsize=36)
    plt.xticks(ticks=[0, max_time / 4, max_time / 2, 3 * max_time / 4, max_time],
               labels=[0, round(max_time / 4), round(max_time / 2), round(3 * max_time / 4), round(max_time)],
               fontsize=36)
    plt.ylabel('ᵠ', fontsize=52, rotation=0, labelpad=10)
    plt.yticks(ticks=[0, np.pi / 2, np.pi, 3 * np.pi / 2, 2 * np.pi], labels=['0', 'π/2', 'π', '3π/2', '2π'],
               fontsize=36)
    plt.hlines([0, np.pi / 2, np.pi, 3 * np.pi / 2, 2 * np.pi], xmin=0, xmax=max_time, linestyles='dashed', linewidth=6)
    plt.axis([0, max_time, 0, 6.3])
    if save_order:
        ordertime = np.array(ordertime)
        np.save(final_directory + '/order_times.npy', ordertime)

        local_order = np.array(local_order)
        locop_plot = plt.figure(3, figsize=(8, 8))
        plt.plot(ordertime, local_order, linewidth=6)
        plt.xlabel('time', fontsize=36)
        plt.ylabel('order parameter', fontsize=36)
        locop_plot.savefig(f"{final_directory}/local_order_parameter.png")
        np.save(final_directory + '/local_order_parameter.npy', local_order)

        global_order = np.array(global_order)
        globop_plot = plt.figure(4, figsize=(18, 18))
        plt.plot(ordertime, global_order, linewidth=12)
        plt.xlabel('time', fontsize=48)
        plt.xticks(ticks=[0, max_time / 4, max_time / 2, 3 * max_time / 4, max_time],
                   labels=[0, round(max_time / 4), round(max_time / 2), round(3 * max_time / 4), round(max_time)],
                   fontsize=48)
        plt.ylabel('order parameter', fontsize=48)
        plt.yticks(ticks=[0, 0.25, 0.451, 0.854, 1], labels=['0', '0.25', '0.45', '0.85', '1'], fontsize=48)
        plt.hlines([0.25, 0.451, 0.854], xmin=0, xmax=max_time, linestyles='dashed', linewidth=6)
        plt.axis([0, max_time, 0, 1.1])
        globop_plot.savefig(f"{final_directory}/global_order_parameter.png")
        np.save(final_directory + '/global_order_parameter.npy', global_order)

    if save_particles or save_twist:
        theta_plot.savefig(f"{final_directory}/theta.png")
        phi_plot.savefig(f"{final_directory}/phi.png")
        param = open(final_directory + '/parameters.txt', 'w')
        param.write(mechanism.parameters)
        param.close()
        times = open(final_directory + '/times_saved.txt', 'w')
        times.write(str(to_plot))
        times.close()
        np.save(final_directory + '/times.npy', x)
        np.save(final_directory + '/theta.npy', y)
        np.save(final_directory + '/phi.npy', z)
    return real_time, theta, phi, slices



import numpy as np
import torch
from pykeops.torch.cluster import sort_clusters
from pykeops.torch import LazyTensor


def volume_ball(d):
    """Volume of the ball of radius 1 in dimension d"""
    if np.remainder(d, 2) == 0:
        V = (np.pi ** (d / 2)) / np.math.factorial(d / 2)
    else:
        k = (d - 1) / 2
        V = 2 * np.math.factorial(k) * (4 * np.pi) ** k / np.math.factorial(d)
    return V


def uniform_grid_separation(x, r, L):
    r"""Return the label of the cell to which each row of **x** belongs.

    The box :math:`[0,L[0]]\times ...\times [0,L[d-1]]` is divided in
    cells of size r. The cells are numbered as follows in dimension 2:

    +--------------------+--------------------+-----+--------------------+
    |``(K[1]-1)*K[0]``   |``(K[1]-1)*K[0]+1`` | ... |``K[0]*K[1]-1``     |
    +--------------------+--------------------+-----+--------------------+
    |...                 |...                 |...  |...                 |
    +--------------------+--------------------+-----+--------------------+
    |``K[0]``            |``K[0]+1``          | ... |``2*K[0]-1``        |
    +--------------------+--------------------+-----+--------------------+
    |``0``               |``1``               | ... |``K[0]-1``          |
    +--------------------+--------------------+-----+--------------------+

    where ``K[i]`` is the number of cells along the dimension ``i``.

    Args:
        x ((N,d) Tensor): Positions.
        r ((d,) Tensor): Size of the cells at each dimension.
        L ((d,) Tensor): Box size.

    Returns:
        (N,) IntTensor:

        The Tensor of labels.

    Raises:
        :exc:`NotImplementedError`: if the dimension is :math:`d=1` or
            :math:`d>3`.

    """
    K = (L / r).floor().int()
    N, d = list(x.size())
    if d == 1 or d > 3:
        raise NotImplementedError()
    index_d = (x / r).floor().int()
    for i in range(d):
        index_d[:, i] = torch.min(index_d[:, i], K[i] - 1)
    if d == 2:
        labels = index_d[:, 0] + K[0] * index_d[:, 1]
    else:
        labels = index_d[:, 0] \
                 + K[0] * index_d[:, 1] \
                 + (K[0] * K[1]) * index_d[:, 2]
    return labels


def uniform_grid_centroids(r, L, d):
    r"""Return the coordinates of the centroids.

    The box :math:`[0,L[0]]\times ...\times [0,L[d-1]]` is divided in
    cells of size r. The cells are numbered as follows in dimension 2:

    +--------------------+--------------------+-----+--------------------+
    |``(K[1]-1)*K[0]``   |``(K[1]-1)*K[0]+1`` | ... |``K[0]*K[1]-1``     |
    +--------------------+--------------------+-----+--------------------+
    |...                 |...                 |...  |...                 |
    +--------------------+--------------------+-----+--------------------+
    |``K[0]``            |``K[0]+1``          | ... |``2*K[0]-1``        |
    +--------------------+--------------------+-----+--------------------+
    |``0``               |``1``               | ... |``K[0]-1``          |
    +--------------------+--------------------+-----+--------------------+

    where ``K[i]`` is the number of cells along the dimension ``i``.

    Args:
        r ((d,) Tensor):  Size of the cells at each dimension.
        L ((d,) Tensor): Box size.
        d (int): Dimension.

    Returns:
        (N_cell,d) Tensor:

        The Tensor of centroids: the :math:`i` component is the center
        of the cell numbered ``i``. The first dimension ``N_cell`` is
        the total number of cells (in the example above,
        ``N_cell = K[0]*K[1]``).

    Raises:
        :exc:`NotImplementedError`: if :math:`d=1` or :math:`d>3`.

    """
    if d == 1 or d > 3:
        raise NotImplementedError()
    K = (L / r).floor().int().type(torch.LongTensor).to(L.device)
    vx = torch.tensor(range(K[0]), dtype=L.dtype, device=L.device)
    vy = torch.tensor(range(K[1]), dtype=L.dtype, device=L.device)
    x = vx.repeat(K[1:].prod())
    x = x.reshape(K.prod(), 1)
    y = vy.repeat_interleave(K[0])
    if d == 3:
        y = y.repeat(K[2])
    y = y.reshape(K.prod(), 1)
    if d == 2:
        centroids = r * torch.cat((x, y), dim=1) + r / 2
    else:
        vz = torch.tensor(range(K[2]), dtype=L.dtype, device=L.device)
        z = vz.repeat_interleave(K[0] * K[1])
        z = z.reshape(K.prod(), 1)
        centroids = r * torch.cat((x, y, z), dim=1) + r / 2
    return centroids


def from_matrix(ranges_i, ranges_j, keep):
    I, J = torch.meshgrid(
        [torch.arange(0, keep.shape[0], device=keep.device),
         torch.arange(0, keep.shape[1], device=keep.device)]
    )
    redranges_i = ranges_i[
        I.t()[keep.t()]
    ]  # Use PyTorch indexing to "stack" copies of ranges_i[...]
    redranges_j = ranges_j[J[keep]]
    slices_i = (
        keep.sum(1).cumsum(0).int()
    )  # slice indices in the "stacked" array redranges_j
    slices_j = (
        keep.sum(0).cumsum(0).int()
    )  # slice indices in the "stacked" array redranges_i
    return (ranges_i, slices_i, redranges_j, ranges_j, slices_j, redranges_i)


def cluster_ranges(lab, Nlab=None):
    if Nlab is None:
        Nlab = torch.bincount(lab).float()
    pivots = torch.cat((torch.tensor([0.0], device=Nlab.device),
                        Nlab.cumsum(0)))
    return torch.stack((pivots[:-1], pivots[1:]), dim=1).int()


def block_sparse_reduction_parameters(x, y, centroids, eps, L, keep,
                                      where_dummies=False):
    """Compute the block sparse reduction parameters.

    Classical block sparse reduction as explained in the documentation
    of the KeOps library. The main difference is that the centroids and
    labels are computed using the :func:`uniform_grid_centroids` and the
    :func:`uniform_grid_separation` respectively. With this method a cell
    may be empty. Dummy particles are thus added to **x** and **y** at
    the positions of the centroids.

    Args:
        x ((M,d) Tensor): Rows.
        y ((N,d) Tensor): Columns
        centroids ((nb_centr,d) Tensor): Centroids.
        eps ((d,) Tensor): Size of the cells.
        L ((d,) Tensor): Size of the box.
        keep ((nb_centro,nb_centro) BoolTensor): ``keep[i,j]`` is True
            when ``centroids[i,:]`` and ``centroids[j,:]`` are
            contiguous.
        where_dummies (bool, optional): Default is False.

    Returns:

        tuple:

        The tuple of block sparse parameters:

        * **x_sorted** `((M+nb_centro,d) Tensor)`: The sorted **x_plus**
          where **x_plus** is the concatenation of **x** and the
          centroids.

        * **y_sorted** `((N+nb_centro,d) Tensor)`: The sorted **y_plus**
          where **y_plus** is the concatenation of **y** and the
          centroids.

        * **nb_centro** `(int)`: Number of centroids.

        * **labels_x** `((M+nb_centro,) IntTensor)`: The labels of
          **x_plus**.

        * **labels_y** `((n+nb_centro,) IntTensor)`: The labels of
          **y_plus**.

        * **permutation_x** `((M+nb_centro,) IntTensor)`: The
          permutation of the **x_plus** such that
          ``x_plus[permutation_x,:]`` is equal to **x_sorted**.

        * **ranges_ij**: Result of :func:`from_matrix`.

        * **map_dummies** `((M+nb_centro,N+nb_centro) LazyTensor)`:
          The :math:`(i,j)` coordinate is 0 if either ``x_sorted[i,:]``
          or ``y_sorted[j,:]`` is a centroid and 1 otherwise.
          **map_dummies** is None when **where_dummies** is False.

    """
    M, D = list(x.shape)
    N, D = list(y.shape)
    # centroids = uniform_grid_centroids(eps, L, D)
    nb_centro, D = list(centroids.size())
    Mplus = M + nb_centro

    x_plus = torch.cat((x, centroids), dim=0)
    y_plus = torch.cat((y, centroids), dim=0)

    labels_x = uniform_grid_separation(x_plus, eps, L)
    ranges_x = cluster_ranges(labels_x)
    labels_y = uniform_grid_separation(y_plus, eps, L)
    ranges_y = cluster_ranges(labels_y)
    x_sorted, labels_x_sorted = sort_clusters(x_plus, labels_x)
    u = torch.arange(0, Mplus, device=x.device).reshape(Mplus, 1)
    permutation_x, label_px = sort_clusters(u, labels_x)
    permutation_x = permutation_x[:, 0]
    y_sorted, labels_y_sorted = sort_clusters(y_plus, labels_y)
    ranges_ij = from_matrix(ranges_x, ranges_y, keep)

    if where_dummies:
        x_dummies = torch.cat(
            (torch.ones(M, dtype=x.dtype, device=x.device),
             torch.zeros(nb_centro, dtype=x.dtype, device=x.device)))
        y_dummies = torch.cat(
            (torch.ones(N, dtype=y.dtype, device=y.device),
             torch.zeros(nb_centro, dtype=y.dtype, device=y.device)))
        x_dummies_sorted, lxd = sort_clusters(x_dummies, labels_x)
        y_dummies_sorted, lyd = sort_clusters(y_dummies, labels_y)
        LTx_dummies = LazyTensor(x_dummies_sorted[:, None, None])
        LTy_dummies = LazyTensor(y_dummies_sorted[None, :, None])
        map_dummies = (LTx_dummies * LTy_dummies).sum(-1)
    else:
        map_dummies = None

    return x_sorted, y_sorted, nb_centro, labels_x, labels_y, \
           permutation_x, ranges_ij, map_dummies


def maximal_eigenvector_dim2(J):
    r"""Eigenvectors of maximum eigenvalue for symmetric matrices in
    dimension 2.

    Args:
        J ((N,2,2) Tensor): Batch of N 2D symmetric matrices.

    Returns:
        (N,2) Tensor:

        Unit eigenvector associated to the maximal eigenvalue of each
        matrix.

    """
    N = J.shape[0]
    a = J[:, 0, 0]
    b = J[:, 0, 1]
    d = J[:, 1, 1]
    sq = (b ** 2) + ((a - d) / 2.) ** 2
    ev = (a + d) / 2. + sq.sqrt()
    u = torch.cat((-b.reshape((N, 1)),
                   (a - ev).reshape((N, 1))
                   ), 1)
    normu = torch.norm(u, dim=1)
    notcool = (normu == 0.)
    nb_notcool = notcool.sum()
    if nb_notcool>0:
        u[notcool, :] = torch.randn((nb_notcool, 2),
                                    dtype=J.dtype,
                                    device=J.device)
    result = u / torch.norm(u, dim=1).reshape((N, 1))
    return result


def quat_mult(q_1, q_2):
    """Multiplication in the space of quaternions."""

    a_1, b_1, c_1, d_1 = q_1[:, 0], q_1[:, 1], q_1[:, 2], q_1[:, 3]
    a_2, b_2, c_2, d_2 = q_2[:, 0], q_2[:, 1], q_2[:, 2], q_2[:, 3]

    q_1_q_2 = torch.stack((
        a_1 * a_2 - b_1 * b_2 - c_1 * c_2 - d_1 * d_2,
        a_1 * b_2 + b_1 * a_2 + c_1 * d_2 - d_1 * c_2,
        a_1 * c_2 - b_1 * d_2 + c_1 * a_2 + d_1 * b_2,
        a_1 * d_2 + b_1 * c_2 - c_1 * b_2 + d_1 * a_2
    ), dim=1)

    return q_1_q_2

def angles_directions_to_quat(angles, directions):
    """Represents a list of rotation angles and axes as quaternions."""
    t = angles / 2
    return torch.cat((t.cos().view(-1, 1),
                      t.sin().view(-1, 1) * directions),
                     dim=1)
